import ast
import inspect
import os
import typing as _typing
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from esrf_pathlib import ESRFPath
from ewoksjob import client

from ..bliss_globals import current_session
from ..import_utils import unavailable_class

try:
    from bliss.common.utils import UserNamespace
except ImportError as ex:
    UserNamespace = unavailable_class(ex)


from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor


class EwoksMacroHandler(
    BaseProcessor, parameters=[ParameterInfo("queue", category="workflows")]
):
    """Ewoks macro's are the equivalent of Bliss macro's with the difference
    that the functions are executed remotely via Ewoks.
    """

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}
        defaults.setdefault("queue", "user")
        super().__init__(config, defaults)

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        scripts = categories.setdefault("Remote scripts", {})
        scripts["Location"] = self.user_script_homedir()
        scripts["#scripts"] = len(self._user_script_list()[1])
        return categories

    def user_script_homedir(self) -> ESRFPath:
        """Root directory for scripts with relative file paths.

        Equivalent of ``user_script_homedir`` from Bliss for local Bliss macro's.
        """
        dataset_filename = ESRFPath(current_session.scan_saving.filename)
        return dataset_filename.scripts_path

    def user_script_load(
        self,
        scriptname: str,
        export_global: Union[str, bool] = "ewoks",
        blocking: bool = False,
        timeout: Optional[float] = None,
    ) -> Optional[UserNamespace]:
        """
        Parse a script to extract function definitions and expose them
        as Ewoks function proxies in the Bliss session's ``env_dict``.

        Each proxy builds a one-node Ewoks workflow dict and submits it.

        Equivalent of ``user_script_load`` from Bliss for local Bliss macro's.

        :param scriptname: The python file to load (absolute path or relative to ``user_script_homedir``).
        :param export_global: Where to install the functions:
                            - ``True``: inject into global session namespace ``env_dict``
                            - "namespace" (`default="ewoks"`): inject into that namespace
                            - ``False``: return a UserNamespace instead of injecting
        :param blocking: Wait for result.
        :param timeout: Timeout when waiting for the result.
        :returns: If ``export_global`` is ``False``, returns a ``UserNamespace`` containing function proxies.
                  Otherwise, functions are injected into the session ``env_dict`` and nothing is returned.
        """
        if not scriptname:
            self._user_script_list()
            return

        filepath = self._resolve_filepath(scriptname)
        function_proxies = self._create_function_proxies(
            filepath, blocking=blocking, timeout=timeout
        )

        if export_global is True:
            current_session._update_env_dict_from_globals_dict(function_proxies)
            return None

        if isinstance(export_global, str):
            ns_name = export_global
            env_dict = current_session.env_dict
            if isinstance(env_dict.get(ns_name), UserNamespace):
                env_dict[ns_name] = env_dict[ns_name] + function_proxies
                print(f"Merged [{ns_name}] namespace in session.")
            else:
                if ns_name in env_dict:
                    print(f"Replaced [{ns_name}] in session env")
                env_dict[ns_name] = UserNamespace(**function_proxies)
                print(f"Exported [{ns_name}] namespace in session.")
            return None

        return UserNamespace(**function_proxies)

    def user_script_list(self):
        """List python scripts from the proposal SCRIPTS directory.

        Equivalent of ``user_script_list`` from Bliss for local Bliss macro's.
        """
        rootdir, files = self._user_script_list()
        print(f"List of python files for remote execution in [{rootdir}]:")
        for scriptname in files:
            print(f" - {scriptname}")

    def _user_script_list(self) -> Tuple[ESRFPath, List[Path]]:
        rootdir = self.user_script_homedir()
        if not rootdir.is_dir():
            return rootdir, []

        files = []
        for dirpath, dirnames, filenames in os.walk(rootdir):
            dirpath = Path(dirpath).relative_to(rootdir)
            for filename in filenames:
                _, ext = os.path.splitext(filename)
                if ext != ".py":
                    continue
                files.append(Path(dirpath) / filename)
        return rootdir, files

    def _resolve_filepath(self, scriptname: str) -> Path:
        """Resolve a scriptname to an absolute Python file path."""
        if not scriptname.endswith(".py"):
            scriptname += ".py"

        if os.path.isabs(scriptname):
            filepath = Path(scriptname)
        else:
            filepath = self.user_script_homedir() / scriptname

        if not filepath.is_file():
            raise RuntimeError(f"Cannot find [{filepath}] !")

        return filepath

    def _create_function_proxies(
        self,
        filepath: Path,
        blocking: bool = False,
        timeout: Optional[float] = None,
    ) -> Dict[str, Callable[..., Any]]:
        """Parse a Python file and return proxies for all public top-level functions."""
        if not filepath.is_absolute():
            raise ValueError(f"File path must be absolute [{filepath}]")

        try:
            with open(filepath, "r", encoding="utf-8") as f:
                source = f.read()
        except Exception as e:
            raise RuntimeError(f"Failed to read file [{filepath}]: {e!r}") from e

        try:
            tree = ast.parse(source, filename=filepath)
        except SyntaxError as e:
            raise RuntimeError(
                f"Syntax error while parsing [{filepath}]: {e.msg} (line {e.lineno})"
            ) from e

        function_proxies = {}

        for node in tree.body:
            if not isinstance(node, ast.FunctionDef) or node.name.startswith("_"):
                continue
            func_name = node.name
            docstring = ast.get_docstring(node)
            signature = _create_function_signature(node)
            function_proxies[func_name] = self._create_function_proxy(
                filepath,
                func_name,
                signature,
                docstring,
                blocking=blocking,
                timeout=timeout,
            )

        return function_proxies

    def _create_function_proxy(
        self,
        filepath: Path,
        func_name: str,
        signature: inspect.Signature,
        doc: Optional[str],
        blocking: bool = False,
        timeout: Optional[float] = None,
    ) -> Callable[..., Any]:
        """Create a function proxy with a specific signature."""

        def wrapper(*args, **kwargs) -> Any:
            remote_function_url = f"{filepath}::{func_name}"
            workflow = {
                "graph": {"id": f"workflow_{func_name}"},
                "nodes": [
                    {
                        "id": "main",
                        "task_type": "method",
                        "task_identifier": remote_function_url,
                    }
                ],
                "links": [],
            }

            inputs = _workflow_inputs_from_signature(signature, args, kwargs)

            try:
                future = client.submit(
                    args=(workflow,), kwargs={"inputs": inputs}, queue=self.queue
                )
            except Exception as exc:
                raise RuntimeError(f"Failed to submit Ewoks job: {exc!r}") from exc

            if not blocking:
                return _Future(future.uuid)
            result = future.result(timeout=timeout)
            return _parse_future_result(result)

        wrapper.__name__ = func_name
        wrapper.__signature__ = signature
        wrapper.__doc__ = (
            doc or f"Ewoks proxy for function {func_name!r} in file {str(filepath)!r}"
        )
        return wrapper


class _Future(client.Future):
    def result(self, *args, **kwargs) -> Any:
        result = super().result(*args, **kwargs)
        return _parse_future_result(result)


def _parse_future_result(result: Any) -> Any:
    if isinstance(result, dict) and "return_value" in result:
        return result["return_value"]
    return result


def _workflow_inputs_from_signature(
    signature: inspect.Signature, args: tuple, kwargs: dict
) -> List[dict]:
    bound = signature.bind_partial(*args, **kwargs)

    # We could fill in the defaults but their evaluation might have failed
    # so let the remote do handle defaults.
    # bound.apply_defaults()

    inputs = []

    has_positional_vars = any(
        p.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.POSITIONAL_ONLY)
        for p in signature.parameters.values()
    )

    idx = 0
    for name, param in signature.parameters.items():
        if name not in bound.arguments:
            continue
        value = bound.arguments[name]

        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            for v in value:
                inputs.append({"name": idx, "value": v})
                idx += 1
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            for k, v in value.items():
                inputs.append({"name": k, "value": v})
        elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
            inputs.append({"name": idx, "value": value})
            idx += 1
        elif param.kind == inspect.Parameter.KEYWORD_ONLY:
            inputs.append({"name": name, "value": value})
        elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            if has_positional_vars:
                inputs.append({"name": idx, "value": value})
                idx += 1
            else:
                inputs.append({"name": name, "value": value})

    return inputs


def _create_function_signature(node: ast.FunctionDef) -> inspect.Signature:
    """Create a function signature from an ``ast.FunctionDef`` node.

    Handles:
      - normal args
      - positional-only args (if present)
      - keyword-only args
      - varargs and kwargs
      - default value fallback
      - annotation eval attempt (safe)
    """
    params: List[inspect.Parameter] = []

    posonly = getattr(node.args, "posonlyargs", []) or []
    regular_args = node.args.args or []
    kwonly_args = node.args.kwonlyargs or []
    defaults = node.args.defaults or []
    kw_defaults = node.args.kw_defaults or []

    # Build defaults aligner for regular args (posonly + regular)
    regular_all = list(posonly) + list(regular_args)
    num_regular = len(regular_all)
    num_defaults = len(defaults)
    defaults_prefix = [None] * (num_regular - num_defaults) + list(defaults)

    # Process positional-only and regular args
    for arg, default_expr in zip(regular_all, defaults_prefix):
        parameter = inspect.Parameter(
            arg.arg,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=_eval_default(default_expr),
            annotation=_parse_annotation(arg.annotation),
        )
        params.append(parameter)

    # varargs (*args)
    if node.args.vararg:
        params.append(
            inspect.Parameter(node.args.vararg.arg, inspect.Parameter.VAR_POSITIONAL)
        )

    # keyword-only args
    for arg, default_expr in zip(kwonly_args, kw_defaults):
        parameter = inspect.Parameter(
            arg.arg,
            inspect.Parameter.KEYWORD_ONLY,
            default=_eval_default(default_expr),
            annotation=_parse_annotation(arg.annotation),
        )
        params.append(parameter)

    # kwargs (**kwargs)
    if node.args.kwarg:
        params.append(
            inspect.Parameter(node.args.kwarg.arg, inspect.Parameter.VAR_KEYWORD)
        )

    return inspect.Signature(params, return_annotation=_parse_annotation(node.returns))


def _parse_annotation(annotation) -> Any:
    if not annotation:
        return inspect.Signature.empty
    try:
        annotation_str = ast.unparse(annotation)
    except Exception:
        return Any
    try:
        return _eval_annotation(annotation_str)
    except Exception:
        # String representation instead of Any so the Bliss REPL completion is informative.
        return annotation_str


def _eval_annotation(annotation_str: str) -> Any:
    safe_builtins = {
        "int": int,
        "float": float,
        "bool": bool,
        "str": str,
        "bytes": bytes,
        "dict": dict,
        "list": list,
        "tuple": tuple,
        "set": set,
        "frozenset": frozenset,
        "object": object,
        "type": type,
        "None": type(None),
    }
    safe_typing = {
        k: getattr(_typing, k) for k in dir(_typing) if not k.startswith("_")
    }
    safe_globals = {
        "__builtins__": {},
        "typing": _typing,
        **safe_builtins,
        **safe_typing,
    }
    return eval(annotation_str, safe_globals, {})


def _eval_default(expr) -> Any:
    if expr is None:
        return inspect.Signature.empty

    try:
        return ast.literal_eval(expr)
    except Exception:
        pass

    try:
        restricted_globals = {"__builtins__": {}}
        return eval(
            compile(ast.Expression(expr), "<ast-default>", "eval"),
            restricted_globals,
            {},
        )
    except Exception:
        pass

    try:
        return ast.unparse(expr)
    except Exception:
        return inspect.Signature.empty



from pathlib import Path


def is_file(filename: str) -> str:
    filepath = Path(filename)
    if not filepath.exists():
        raise ValueError(f"File {filename} does not exist")
    if not filepath.is_file():
        raise ValueError(f"{filename} is not a file")
    return str(filepath)



import getpass
import os
import shutil
from typing import Optional

from ....resources import resource_path
from . import defaults


def raw_directory(session: str, sample: str, dataset: str, demo: bool = False) -> str:
    path = normalize_path(
        session, "RAW_DATA", sample, f"{sample}_{dataset}", f"{sample}_{dataset}.h5"
    )
    if demo and path.startswith(defaults.REAL_DATA_ROOT):
        demo_path = path.replace(defaults.REAL_DATA_ROOT, defaults.DEMO_DATA_ROOT)
        if os.path.exists(demo_path):
            path = demo_path
    return path


def pymca_config_path(session: str, name: str, demo: bool = False) -> str:
    if os.path.isabs(name):
        return name
    path = normalize_path(session, "SCRIPTS")
    if demo and path.startswith(defaults.REAL_DATA_ROOT):
        demo_path = path.replace(defaults.REAL_DATA_ROOT, defaults.DEMO_DATA_ROOT)
        if os.path.exists(demo_path):
            path = demo_path
    return os.path.join(path, "pymca", name)


def processed_path(
    session: str,
    sample: str,
    dataset: str,
    dirname: Optional[str],
    outname: str,
    demo: bool = False,
) -> str:
    if dirname is None:
        dirname = defaults.DEFAULT_OUT_DIRNAME
    path = normalize_path(session, "PROCESSED_DATA", dirname, outname)
    return _change_demo_path(session, sample, dataset, path, demo=demo)


def accessible_workflow_path(
    session: str, sample: str, dataset: str, dirname: Optional[str], demo: bool = False
) -> str:
    if dirname is None:
        dirname = defaults.DEFAULT_OUT_DIRNAME
    path = normalize_path(session, "SCRIPTS", dirname, "workflows")
    return _change_demo_path(session, sample, dataset, path, demo=demo)


def ensure_workflow_exists(path: str) -> None:
    if os.path.exists(path):
        return
    workflow_basename = os.path.basename(path)
    with resource_path("fluo", workflow_basename) as src_path:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        shutil.copyfile(src_path, path)


def _change_demo_path(
    session: str, sample: str, dataset: str, path: str, demo: bool = False
) -> str:
    raw_path = raw_directory(session, sample, dataset, demo=demo)
    if not demo or not raw_path.startswith(defaults.DEMO_DATA_ROOT):
        return path

    job_id = os.environ.get("SLURM_JOBID")
    if job_id:
        out_root = f"/tmp_14_days/{getpass.getuser()}/ewoksdemo/slurm_job_{job_id}"
    else:
        out_root = "/tmp/ewoksdemo"
    return path.replace(defaults.REAL_DATA_ROOT, out_root)


def normalize_path(path: str, *parts) -> str:
    if parts:
        path = os.path.join(path, *parts)
    path = os.path.abspath(path)
    path = path.replace("/mnt/multipath-shares", "")
    return path



import logging
import os
import re
import shutil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from ewoksjob.client import get_future
from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import directories
from .parameters.fluoxas import fluoxas_workflow_inputs
from .parameters.mosaic_xrfmap import mosaic_xrfmap_workflow_inputs
from .parameters.xrfmap import xrfmap_workflow_inputs

logger = logging.getLogger(__name__)


class FluoProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("queue", category="workflows"),
        ParameterInfo("xrf_names", category="data access"),
        ParameterInfo("virtual_axes", category="regrid"),
        ParameterInfo("ignore_axes", category="regrid"),
        ParameterInfo("pymca_configs", category="PyMCA"),
        ParameterInfo("energy_name", category="PyMCA"),
        ParameterInfo("quantification", category="PyMCA"),
        ParameterInfo("norm_counter_name", category="Normalization"),
        ParameterInfo("data_portal_upload", category="data portal"),
    ],
):
    """A class that holds parameters related to online workflow triggering for PyMCA."""

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("workflow", "")
        defaults.setdefault("xrf_names", [])
        defaults.setdefault("quantification", True)
        defaults.setdefault("virtual_axes", {})
        defaults.setdefault("ignore_axes", [])
        defaults.setdefault("data_portal_upload", False)
        defaults.setdefault("trigger_at", "END")

        super().__init__(config=config, defaults=defaults)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[dict]:
        metadata, _ = self._on_new_scan(scan)
        return metadata

    def _on_new_scan(self, scan: BlissScanType) -> Tuple[Optional[dict], Optional[Any]]:
        """Executed at the end of every scan"""
        future = None
        metadata = None
        if not self.scan_requires_processing(scan):
            return metadata, future

        upload_parameters = None
        if self.data_portal_upload and scan.scan_info.get("save"):
            # It is allowed to upload the same processed directory more than once.
            upload_parameters = self._get_workflow_upload_parameters(scan)

        # Submit arguments
        workflow_base, inputs = self.get_workflow_and_inputs(scan)
        workflow = self.get_workflow_dst(scan, workflow_base)
        kwargs = self.get_submit_arguments(scan, inputs)
        if scan.scan_info.get("save"):
            kwargs["convert_destination"] = directories.workflow_destination(scan)
        if upload_parameters:
            kwargs["upload_parameters"] = upload_parameters

        # Trigger workflow from the current process.
        print("Starting workflow", workflow, kwargs)
        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        future = get_future(future.uuid)

        return metadata, future

    def get_submit_arguments(self, scan: BlissScanType, inputs: List[dict]) -> dict:
        return {
            "inputs": inputs,
            "outputs": [{"all": False}],
        }

    def _set_workflow(self, scan: BlissScanType, filename) -> None:
        """Set the workflow filename for the scan"""
        self.workflow = filename

    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def _get_axis_units_from_scan(
        self, virtual_axes: Dict[str, str], scan: BlissScanType
    ):
        units = scan.scan_info["positioners"]["positioners_units"]
        units_dict = {}
        for virtual_axis in virtual_axes:
            for motor_name in _get_motors_from_expression(virtual_axis):
                motor_unit = units[motor_name]
                if motor_unit is None:
                    logger.warning(
                        f"Unit for motor `{motor_name}` is not set in beacon!"
                    )
                else:
                    units_dict[motor_name] = motor_unit

        return units_dict

    def get_workflow_and_inputs(self, scan: BlissScanType):
        scan_info = scan.scan_info

        if scan_info.get("is_xrfmap_patch"):
            concat_filename = directories.master_output_filename(scan).replace(
                ".h5", "_concat.h5"
            )
            scan_ranges = _get_scan_ranges_from_sequence(scan)

            workflow, inputs = mosaic_xrfmap_workflow_inputs(
                filenames=[self.get_filename(scan)],
                output_root_uri=self.master_output_url(scan),
                scan_ranges=[scan_ranges],
                concat_bliss_scan_uri=f"{concat_filename}::/{scan_ranges[0]}.1",
                config_filenames=self.pymca_configs,
                detector_names=bliss_counter_to_h5_name(self.xrf_names),
                counter_name=self.norm_counter_name,
                energy_name=self.energy_name,
                quantification=self.quantification,
                virtual_axes=self.virtual_axes,
                ignore_axes=self.ignore_axes,
                axis_units=self._get_axis_units_from_scan(
                    self.virtual_axes.values(), scan
                ),
            )
        elif scan_info.get("is_fluoxas"):
            workflow, inputs = fluoxas_workflow_inputs(
                filenames=[self.get_filename(scan)],
                output_root_uri=self.master_output_url(scan),
                scan_ranges=[_get_scan_ranges_from_sequence(scan)],
                config_filenames=self.pymca_configs,
                detector_names=bliss_counter_to_h5_name(self.xrf_names),
                counter_name=self.norm_counter_name,
                energy_name=self.energy_name,
                quantification=self.quantification,
                virtual_axes=self.virtual_axes,
                ignore_axes=self.ignore_axes,
                axis_units=self._get_axis_units_from_scan(
                    self.virtual_axes.values(), scan
                ),
            )
        else:
            workflow, inputs = xrfmap_workflow_inputs(
                filename=self.get_filename(scan),
                output_root_uri=self.master_output_url(scan),
                scan_number=scan_info.get("scan_nb"),
                config_filenames=self.pymca_configs,
                detector_names=bliss_counter_to_h5_name(self.xrf_names),
                counter_name=self.norm_counter_name,
                energy_name=self.energy_name,
                quantification=self.quantification,
                virtual_axes=self.virtual_axes,
                ignore_axes=self.ignore_axes,
                axis_units=self._get_axis_units_from_scan(
                    self.virtual_axes.values(), scan
                ),
            )
        return workflow + ".ows", inputs

    def get_workflow_dst(
        self, scan: BlissScanType, workflow_base: str
    ) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        src_file = resource_filename("fluo", workflow_base)
        if src_file is None:
            return

        dataset_filename = self.get_filename(scan)
        workflow_directory = self._get_workflows_dir(dataset_filename)
        dst_file = os.path.join(workflow_directory, os.path.basename(src_file))
        if src_file != dst_file:
            self._set_workflow(scan, dst_file)

        if not os.path.exists(dst_file):
            os.makedirs(workflow_directory, exist_ok=True)
            shutil.copyfile(src_file, dst_file)

        return dst_file

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        return directories.get_workflows_dir(dataset_filename)

    def add_xrf_names(self, detector, *spectra: Sequence[int]) -> None:
        """Add an xrf detector and its spectra to the processor"""
        xrf_names = set(self.xrf_names)
        for counter in detector.counters:
            if counter.name.startswith("spectrum_det"):
                spectrum_id = int(counter.name.replace("spectrum_det", ""))
                if spectra is None or spectrum_id in spectra:
                    xrf_names.add(counter.fullname)
        self.xrf_names = sorted(xrf_names)

    def clear_xrf_names(self):
        self.xrf_names = []

    def get_xrf_names(self, scan: BlissScanType) -> List[str]:
        # If this is a sequence check the first real scan
        if scan.scan_info.get("is_scan_sequence"):
            channels = scan.streams["SUBSCANS"][0].info.get("channels", dict())
        else:
            channels = scan.scan_info.get("channels", dict())

        return sorted((xrf_name for xrf_name in self.xrf_names if xrf_name in channels))

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        scan_info = scan.scan_info
        scan_to_process = (
            scan_info.get("is_xrfmap")
            or scan_info.get("is_xrfmap_patch")
            or scan_info.get("is_fluoxas")
        )
        return (
            bool(self.get_xrf_names(scan))
            and bool(scan.scan_info.get("save"))
            and bool(scan_to_process)
        )

    def master_output_url(self, scan: BlissScanType) -> str:
        """URL which can be used to inspect the results after the processing."""
        scan_nb = scan.scan_info.get("scan_nb")
        filename = directories.master_output_filename(scan)
        return f"{filename}::/{scan_nb}.1"

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> dict:
        raw_directory = os.path.dirname(self.get_filename(scan))
        processed_directory = directories.scan_processed_directory(scan)
        scan_saving = current_session.scan_saving
        metadata = {"Sample_name": scan_saving.dataset["Sample_name"]}

        return {
            "beamline": scan_saving.beamline,
            "proposal": scan_saving.proposal_name,
            "dataset": "integrate",
            "path": processed_directory,
            "raw": [raw_directory],
            "metadata": metadata,
        }


def bliss_counter_to_h5_name(names: List[str]):
    """Mangle bliss counter to h5 dataset name"""
    h5_names = []
    for name in names:
        h5_names.append(name.replace(":", "_").replace("spectrum_det", "det"))
    return h5_names


def _get_motors_from_expression(
    expression: str,
    start_var: str = "<",
    end_var: str = ">",
) -> Tuple[str, str]:
    pattern = rf"{re.escape(start_var)}([^{re.escape(end_var)}]+){re.escape(end_var)}"
    return re.findall(pattern, expression)


def _get_scan_ranges_from_sequence(scan: BlissScanType):
    final_scan = scan.streams["SUBSCANS"][:][-1]
    return (2, final_scan.info["scan_nb"])



"""User API for HDF5 conversion on the Bliss repl"""

import os
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename


class Id12Hdf5ToAsciiConverter(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("external_proposal_outdir", category="parameters"),
        ParameterInfo("inhouse_proposal_outdir", category="parameters"),
        ParameterInfo("counters", category="parameters"),
        ParameterInfo("test_proposal_outdir", category="parameters"),
        ParameterInfo("retry_timeout", category="data access"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("trigger_at", "END")
        root_dir = "/data/id12/inhouse"
        defaults.setdefault(
            "external_proposal_outdir", os.path.join(root_dir, "EXTERNAL")
        )
        defaults.setdefault(
            "inhouse_proposal_outdir", os.path.join(root_dir, "INHOUSE2")
        )
        defaults.setdefault("counters", "all")
        defaults.setdefault("test_proposal_outdir", os.path.join(root_dir, "NOBACKUP"))
        defaults.setdefault("workflow", resource_filename("id12", "convert.json"))

        super().__init__(config=config, defaults=defaults)

        # For integration tests
        self._future = None

    def on_new_scan_metadata(self, scan: BlissScanType) -> None:
        if not self.scan_requires_processing(scan):
            return
        kwargs = self.get_submit_arguments(scan)
        self._future = submit(args=(self.workflow,), kwargs=kwargs)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.on_new_scan_metadata(scan)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        return scan.scan_info["filename"] and scan.scan_info["save"]

    def get_submit_arguments(self, scan: BlissScanType) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_inputs(self, scan: BlissScanType) -> List[dict]:
        task_identifier = "Hdf5ToAscii"

        filename = scan.scan_info["filename"]
        output_dir = self.output_dir(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        has_subscan = len(scan.acq_chain.tree.children("root")) == 2

        return [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_dir",
                "value": output_dir,
            },
            {
                "task_identifier": task_identifier,
                "name": "scan_numbers",
                "value": [scan_nb],
            },
            {
                "task_identifier": task_identifier,
                "name": "counters",
                "value": self.counters or "all",
            },
            {
                "task_identifier": task_identifier,
                "name": "has_subscan",
                "value": has_subscan,
            },
        ]

    def output_dir(self, scan: BlissScanType) -> str:
        proposal = current_session.scan_saving.proposal.name

        # Proposal directory is upper case and "-" between letters and digits
        matches = re.findall(r"[A-Za-z]+|\d+", proposal.upper())
        proposal_dir = "-".join(matches)

        # Handle special cases
        proposal_dir = proposal_dir.replace("IH", "IH-")
        proposal_dir = re.sub(r"ID-(\d{2})(.+)", r"ID\1-\2", proposal_dir)

        # Select directory for the ASCII files
        if current_session.scan_saving.proposal_type == "inhouse":
            dirname = self.inhouse_proposal_outdir
        elif current_session.scan_saving.proposal_type == "tmp":
            dirname = self.test_proposal_outdir
        elif proposal_dir.startswith("IH") or proposal_dir.startswith("BLC"):
            dirname = self.inhouse_proposal_outdir
        else:
            dirname = self.external_proposal_outdir

        filename = scan.scan_info["filename"]
        collection_dataset = os.path.splitext(os.path.basename(filename))[0]
        return os.path.join(dirname, proposal_dir, collection_dataset)



from importlib.metadata import version

from blissdata.beacon.data import BeaconData
from packaging.version import Version

_BLISSDATA_VERSION = Version(version("blissdata"))

# bliss 1.11 -> blissdata 0.3.x
# bliss 2.0  -> blissdata 1.0.x
# bliss 2.1  -> blissdata 1.1.x
# master     -> blissdata 2.0.x

if _BLISSDATA_VERSION >= Version("1.1"):
    from pydantic import Field  # noqa
else:
    from pydantic.v1 import Field  # noqa


def get_redis_db_url():
    if _BLISSDATA_VERSION >= Version("1.0"):
        return BeaconData().get_redis_db()

    raw_url = BeaconData().get_redis_db()
    _, url = raw_url.split(":")

    if url.endswith("sock"):
        return f"unix://{url}"
    else:
        return f"redis://{url}"



import logging
from functools import wraps
from importlib.metadata import version
from typing import Callable

from packaging.version import Version

from ..import_utils import unavailable_class

try:
    if Version(version("bliss")) >= Version("2.2"):
        from flint.client.plots import BasePlot
        from flint.client.proxy import FlintClient
    else:
        from bliss.flint.client.plots import BasePlot
        from bliss.flint.client.proxy import FlintClient
except ImportError as ex:
    FlintClient = unavailable_class(ex)
    BasePlot = unavailable_class(ex)

logger = logging.getLogger(__name__)


def capture_errors(method) -> Callable:
    @wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            msg = f"Flint plot error: {e}"
            logger.error(msg, exc_info=True)

    return wrapper



from ewoksjob.client import Future

from ..flint.plotter import BasePlotter
from .plots import BlissIntensityPlots
from .plots import BlissOffsetPlot


class CdiPlotter(BasePlotter):

    def handle_workflow_result(
        self, future: Future, retry_timeout: int, retry_period: int
    ):
        self._spawn(self._handle_workflow_result, future, retry_timeout, retry_period)

    def _handle_workflow_result(
        self, future: Future, retry_timeout: int, retry_period: int
    ):
        result = future.result(timeout=retry_timeout, interval=retry_period)

        output_url = result["saved_hdf5"]

        self._get_plot("IntensityPlots", BlissIntensityPlots).set_data(output_url)
        self._get_plot("OffsetPlot", BlissOffsetPlot).set_data(output_url)



import logging
import os
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import Tuple

import numpy
from silx.io import h5py_utils
from silx.utils.retry import RetryError
from silx.utils.retry import RetryTimeoutError

from .models import XrpdFieldName
from .models import XrpdPlotInfo
from .utils import get_axis_data
from .utils import nfs_cache_refresh

logger = logging.getLogger(__name__)


def create_plot_info(
    scan_name: str,
    lima_name: str,
    radial_label: str,
    azim_label: Optional[str],
    plot_data: Dict[XrpdFieldName, numpy.ndarray],
    hdf5_url: Optional[str] = None,
) -> XrpdPlotInfo:
    plot_info = XrpdPlotInfo(
        scan_name=scan_name,
        lima_name=lima_name,
        radial_label=radial_label,
        azim_label=azim_label,
        hdf5_url=hdf5_url,
        field_names=list(plot_data.keys()),
    )
    plot_info.save()
    logger.debug("XrpdPlotInfo CREATE %r", plot_info.legend)
    plot_info.delete_data_arrays()
    for key, data in plot_data.items():
        persistent_array = plot_info.get_data_array(key)
        persistent_array.extend(data)
    return plot_info


def get_plots_to_remove(max_len: int) -> List[XrpdPlotInfo]:
    all_plots = get_plots()

    if max_len <= 0:
        return all_plots

    remove: List[XrpdPlotInfo] = list()
    keep_scans: Set[str] = set()
    for plot_info in all_plots[::-1]:
        scan_name = plot_info.scan_name
        if len(keep_scans) == max_len and scan_name not in keep_scans:
            remove.append(plot_info)
        else:
            keep_scans.add(scan_name)

    return remove


def delete_plot_info(plot_info: XrpdPlotInfo) -> None:
    plot_info.delete_data_arrays()
    XrpdPlotInfo.delete(plot_info.pk)
    logger.debug("XrpdPlotInfo DEL %r", plot_info.legend)


def get_curve_data(
    plot_key: str, **retry_options
) -> Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray], XrpdPlotInfo]:
    """Get the data from the results of the last processed point of the scan."""
    plot_info = XrpdPlotInfo.get(plot_key)

    try:
        x = plot_info.get_data_array("radial")[()]
    except IndexError:
        return None, None, plot_info

    y = _get_last_integrated_intensity(plot_info, **retry_options)
    return x, y, plot_info


def append_image_data(
    plot_key: str,
    current_data: Optional[numpy.ndarray] = None,
    **retry_options,
) -> Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray], XrpdPlotInfo]:
    """Add new data (if any) to the current data (if any)"""
    plot_info = XrpdPlotInfo.get(plot_key)
    try:
        x = plot_info.get_data_array("radial")[()]
    except IndexError:
        return None, None, plot_info

    if plot_info.hdf5_url:
        if current_data is None:
            idx = tuple()
        else:
            idx = slice(len(current_data), None)
        try:
            y = _get_data_from_file(plot_info.hdf5_url, idx=idx, **retry_options)
        except RetryTimeoutError:
            y = None
        else:
            if current_data is not None:
                y = numpy.vstack([current_data, y])
    else:
        y = plot_info.get_data_array("intensity")[()]

    return x, y, plot_info


def _get_last_integrated_intensity(plot_info: XrpdPlotInfo, **retry_options):
    if plot_info.hdf5_url:
        try:
            return _get_data_from_file(plot_info.hdf5_url, idx=-1, **retry_options)
        except RetryTimeoutError:
            return None

    try:
        return plot_info.get_data_array("intensity")[-1]
    except IndexError:
        return None


def get_2d_integration_data(plot_key: str, **retry_options) -> Tuple[
    Optional[numpy.ndarray],
    Optional[numpy.ndarray],
    Optional[numpy.ndarray],
    XrpdPlotInfo,
]:
    plot_info = XrpdPlotInfo.get(plot_key)

    try:
        x = plot_info.get_data_array("radial")[()]
        y = plot_info.get_data_array("azimuthal")[()]
    except IndexError:
        return None, None, None, plot_info

    intensity = _get_last_integrated_intensity(plot_info, **retry_options)
    return x, y, intensity, plot_info


@h5py_utils.retry()
def _get_data_from_file(hdf5_url: str, idx=tuple()):
    filename, dsetname = hdf5_url.split("::")
    nfs_cache_refresh(os.path.dirname(filename))
    with h5py_utils.File(filename) as root:
        try:
            return root[dsetname][idx]
        except KeyError as e:
            raise RetryError(str(e))


def get_plots() -> List[XrpdPlotInfo]:
    return sorted(
        [XrpdPlotInfo.get(pk) for pk in XrpdPlotInfo.all_pks()],
        key=lambda p: p.timestamp,
    )


def delete_old_entries(current_session) -> int:
    """Delete Redis entries from an older blissoda version"""
    db = XrpdPlotInfo._meta.database
    old_keys = db.keys(f"blissoda:{current_session.name}:Plotter*")
    for key in old_keys:
        XrpdPlotInfo._meta.database.delete(key)
    return len(old_keys)


class XrpdNXDataInfo(NamedTuple):
    radial_label: str
    azim_label: Optional[str]
    plot_data: Dict[XrpdFieldName, numpy.ndarray]
    intensity_url: str


@h5py_utils.retry()
def get_xrpd_nxdata_info(nxdata_url: str) -> XrpdNXDataInfo:
    """Returns NXdata radial axis info and HDF5 URL of the intensity dataset"""
    plot_data: Dict[XrpdFieldName, numpy.ndarray] = dict()

    filename, nxdata_name = nxdata_url.split("::")
    nfs_cache_refresh(os.path.dirname(filename))
    with h5py_utils.File(filename, mode="r") as f:
        try:
            nxdata = f[nxdata_name]
            radial_name = nxdata.attrs["axes"][-1]
            radial_label, radial_data = get_axis_data(nxdata, radial_name)
            plot_data["radial"] = radial_data

            signal = nxdata[nxdata.attrs["signal"]]
            if signal.ndim > 2:
                azim_name = nxdata.attrs["axes"][-2]
                azim_label, azim_data = get_axis_data(nxdata, azim_name)
                plot_data["azimuthal"] = azim_data
            else:
                azim_label = None
        except KeyError as e:
            raise RetryError(str(e))

        return XrpdNXDataInfo(
            radial_label=radial_label,
            azim_label=azim_label,
            plot_data=plot_data,
            intensity_url=f"{filename}::{signal.name}",
        )



import os
from typing import Tuple

from ..bliss_globals import current_session


def get_dataset_dir(dataset_filename: str) -> str:
    return os.path.dirname(os.path.abspath(dataset_filename))


def get_collection_dir(dataset_filename: str) -> str:
    return _abs_join(get_dataset_dir(dataset_filename), "..")


def get_raw_dir(dataset_filename: str) -> str:
    return _abs_join(get_collection_dir(dataset_filename), "..")


def get_proposal_dir(dataset_filename: str) -> str:
    dirname = get_raw_dir(dataset_filename)
    if os.path.basename(dirname) == "raw":
        # version 2
        return _abs_join(dirname, "..")
    # version 1: proposal == raw
    return dirname


def get_processed_dir(dataset_filename: str) -> str:
    version, dirname = get_directory_version(dataset_filename)
    if version == 3:
        return _abs_join(dirname, "..", "PROCESSED_DATA")
    if version == 2:
        return _abs_join(dirname, "..", "processed")
    return _abs_join(dirname, "processed")


def get_dataset_processed_dir(dataset_filename: str, *subdirs) -> str:
    root = get_processed_dir(dataset_filename, *subdirs)
    collection = os.path.basename(get_collection_dir(dataset_filename))
    dataset = os.path.basename(get_dataset_dir(dataset_filename))
    return os.path.join(root, collection, dataset)


def get_processed_subdir(dataset_filename, *subdirs) -> str:
    return os.path.join(get_processed_dir(dataset_filename), *subdirs)


def get_workflows_dir(dataset_filename: str) -> str:
    return get_processed_subdir(dataset_filename, "workflows")


def get_nobackup_dir(dataset_filename: str) -> str:
    version, dirname = get_directory_version(dataset_filename)
    if version == 3:
        return _abs_join(dirname, "..", "NOBACKUP")
    if version == 2:
        return _abs_join(dirname, "..", "_nobackup")
    return _abs_join(dirname, "_nobackup")


def get_directory_version(dataset_filename: str) -> Tuple[int, str]:
    """Returns directory structure version number and the raw data directory"""
    dirname = get_raw_dir(dataset_filename)
    if os.path.basename(dirname) == "RAW_DATA":
        return 3, dirname
    if os.path.basename(dirname) == "raw":
        return 2, dirname
    # proposal == raw
    return 1, dirname


def get_session_dir(dataset_filename: str) -> str:
    return _abs_join(get_proposal_dir(dataset_filename), "..")


def _abs_join(*args):
    return os.path.abspath(os.path.join(*args))


def get_filename(scan) -> str:
    filename = scan.scan_info.get("filename")
    if filename:
        return filename
    return current_session.scan_saving.filename


def scan_processed_directory(scan) -> str:
    return get_dataset_processed_dir(get_filename(scan))


def workflow_destination(scan) -> str:
    filename = get_filename(scan)
    scan_nb = scan.scan_info.get("scan_nb")
    root = scan_processed_directory(scan)
    stem = os.path.splitext(os.path.basename(filename))[0]
    basename = f"{stem}_{scan_nb:04d}.json"
    return os.path.join(root, basename)


def master_output_filename(scan) -> str:
    """Filename which can be used to inspect the results after the processing."""
    filename = get_filename(scan)
    root = scan_processed_directory(scan)
    basename = os.path.basename(filename)
    return os.path.join(root, basename)



"""This module provides pytest-like features for integration tests while
preserving the capability to run the tests manualy from the Bliss shell.

The difference with pytest:

- Fixture functions cannot have the same name, even when defined in a different module.
- Fixtures always have the 'function' scope.
- Tests are not auto-discovered but need the ``@integration_fixture`` decorator
  to ensure the fixture arguments get evaluated.

A minimal example:

.. code-block: python

    from blissoda.demo import testing

    @testing.integration_fixture(autouse=True)
    def _example_auto_fixture():
        print("'_example_auto_fixture' is always called when imported")

    @testing.integration_fixture
    def _example_fixture1():
        print("setup '_example_fixture1'")
        yield "value1"
        print("teardown '_example_fixture1' always called, also when test fails")

    @testing.integration_fixture
    def _example_fixture2(_example_fixture1):
        print("execute '_example_fixture2'")
        assert _example_fixture1 == "value1"
        return "value2"

    @testing.integration_test
    def example_test(_example_fixture1, _example_fixture2, expo=0.2, npoints=10):
        assert _example_fixture1 == "value1"
        assert _example_fixture2 == "value2"

The fixtures are private to hide them from the Bliss shell.
"""

from . import _auto  # noqa F401: ensure that autouse fixtures are imported
from ._assert import demo_assert  # noqa F401
from ._celery import wait_workflows  # noqa F401
from ._fixtures import integration_fixture  # noqa F401
from ._hdf5 import assert_hdf5_dataset_exists  # noqa F401
from ._spec import assert_spec_scan_exists  # noqa F401
from ._test import integration_test  # noqa F401



"""Integration tests for Ewoks macros."""

import inspect
import typing

from .. import testing
from ..processors.ewoks_macros import DemoEwoksMacroHandler

EWOKS_MACROS = DemoEwoksMacroHandler()


def ewoks_macro_demo():
    test_ewoks_macro_list()
    test_ewoks_macro_load_and_execute()
    test_ewoks_macro_arguments()


@testing.integration_fixture
def ewoks_macros():
    rootdir = EWOKS_MACROS.user_script_homedir()
    file1 = rootdir / "file1.py"
    file2 = rootdir / "subdir1" / "file2.py"
    file3 = rootdir / "file3.py"

    file2.parent.mkdir(parents=True, exist_ok=True)

    file1.write_text(_REMOTE_FILE1)
    file2.write_text(_REMOTE_FILE2)
    file3.write_text(_REMOTE_FILE3)

    yield EWOKS_MACROS


_REMOTE_FILE1 = '''
def func_with_docstring(n):
    """Doc string func_with_docstring"""
    return list(range(n))
'''

_REMOTE_FILE2 = """
from typing import List

def func_with_annotations(n:int=4) -> List:
    return list(range(n))
"""


_REMOTE_FILE3 = """
from typing import List, Tuple, Optional, Any, Dict

def func_many_types(
    a: int,
    b: str,
    /,                  # <-- positional-only up to here
    c: bool,
    d: float,
    *args: int,         # <-- after this, only keyword args allowed
    e: Tuple[int] = (1, 2, 3),
    f: Optional[str] = None,
    g: List[int] = [42],
    **kwargs: int
) -> Dict[str, Any]:
    return {
        'a': a,
        'b': b,
        'c': c,
        'd': d,
        'e': e,
        'f': f,
        'g': g,
        'args': args,
        'kwargs': kwargs
    }
"""


@testing.integration_test
def test_ewoks_macro_list(ewoks_macros):
    _, files = ewoks_macros._user_script_list()
    base_names = {filepath.name for filepath in files}
    expected = {"file1.py", "file2.py", "file3.py"}
    assert base_names == expected, base_names


@testing.integration_test
def test_ewoks_macro_load_and_execute(ewoks_macros):
    ns1 = ewoks_macros.user_script_load(
        "file1.py", export_global=False, blocking=True, timeout=10
    )
    result1 = ns1.func_with_docstring(3)
    assert result1 == [0, 1, 2], result1

    ns2 = ewoks_macros.user_script_load(
        "subdir1/file2.py", export_global=False, blocking=True, timeout=10
    )
    result2 = ns2.func_with_annotations()
    assert result2 == [0, 1, 2, 3], result2

    ns1.func_with_docstring.__doc__ == "Doc string func_with_docstring"
    ns2.func_with_annotations.__doc__.startswith(
        "Ewoks proxy for function 'func_with_annotations' in file '"
    )
    ns2.func_with_annotations.__doc__.endswith("file2.py'")


@testing.integration_test
def test_ewoks_macro_arguments(ewoks_macros):
    ns = ewoks_macros.user_script_load(
        "file3.py", export_global=False, blocking=True, timeout=10
    )

    args = ([-1], -2)
    kwargs = {"extra": -3}
    result = ns.func_many_types(
        42, "hello", False, 2.71, *args, f="world", g=[4, 5, 6], **kwargs
    )

    expected = {
        "a": 42,
        "b": "hello",
        "c": False,
        "d": 2.71,
        "e": (1, 2, 3),
        "f": "world",
        "g": [4, 5, 6],
        "args": args,
        "kwargs": kwargs,
    }
    assert result == expected, result

    sig = inspect.signature(ns.func_many_types)
    param_a = sig.parameters["a"]
    param_e = sig.parameters["e"]
    param_g = sig.parameters["g"]
    type_e = param_e.annotation.__origin__
    type_g = param_g.annotation.__origin__
    type_return = sig.return_annotation.__origin__
    assert param_a.annotation is int, param_a.annotation
    assert param_e.default == (1, 2, 3), param_e.default
    assert type_e in (tuple, typing.Tuple), type_e
    assert type_g in (list, typing.List), type_g
    assert type_return in (dict, typing.Dict), type_return



"""Send a workflow to the ewoks Celery application and recieve the
intermediate results (ewoks events) or final result (job return value)
"""

import os
import tempfile

from ewoksjob.client import submit
from ewoksjob.events.readers import instantiate_reader

from . import EWOKS_EVENTS_DIR


def example_job(redis=False):
    # Events during execution
    if redis:
        # Redis backend
        events_url = "redis://localhost:10002/5"
        handlers = [
            {
                "class": "ewoksjob.events.handlers.RedisEwoksEventHandler",
                "arguments": [{"name": "url", "value": events_url}],
            }
        ]
    else:
        # SQLite backend
        events_url = f"file://{EWOKS_EVENTS_DIR}"
        handlers = [
            {
                "class": "ewoksjob.events.handlers.Sqlite3EwoksEventHandler",
                "arguments": [{"name": "uri", "value": events_url}],
            }
        ]
    print("Ewoks events:", events_url)

    # Test workflow to execute
    workflow = {
        "graph": {"id": "mygraph"},
        "nodes": [
            {"id": "task1", "task_type": "method", "task_identifier": "numpy.add"},
            {"id": "task2", "task_type": "method", "task_identifier": "numpy.add"},
        ],
        "links": [
            {
                "source": "task1",
                "target": "task2",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            }
        ],
    }

    # Job arguments
    out_dir = tempfile.mkdtemp(dir=os.environ["DEMO_TMP_ROOT"], prefix="test_job_")
    print("Artifacts:", out_dir)
    varinfo = {"root_uri": out_dir, "scheme": "nexus"}
    inputs = [
        {"id": "task1", "name": 0, "value": 1},
        {"id": "task1", "name": 1, "value": 2},
        {"id": "task2", "name": 1, "value": 3},
    ]
    execinfo = {"handlers": handlers}
    args = (workflow,)
    kwargs = {
        "engine": None,
        "execinfo": execinfo,
        "inputs": inputs,
        "varinfo": varinfo,
        "outputs": [{"all": False}],
    }

    # Trigger workflow
    future = submit(args=args, kwargs=kwargs)
    job_id = future.uuid

    # events could be received in the mean time (see below)
    print(f"Wait for job {job_id} to finished ...")
    workflow_results = future.result(timeout=3, interval=0.1)
    print("Results:", workflow_results)
    assert workflow_results == {"return_value": 6}

    reader = instantiate_reader(events_url)

    # Get intermediate results from ewoks events
    results_during_execution = list(reader.get_events(job_id=job_id))
    assert len(results_during_execution) == 8  # start/stop for job, workflow and node

    try:
        # Get start event of node "task1"
        result_event = list(
            reader.get_events_with_variables(
                job_id=job_id, node_id="task1", type="start"
            )
        )
    except ImportError as e:
        print(f"Cannot read ewoks event URL's: {e}")
    else:
        assert len(result_event) == 1
        result_event = result_event[0]

        # Get access to all output variables of "task1"
        if varinfo.get("root_uri"):
            results = result_event["outputs"]
            assert results["return_value"].value == 3



import logging
import os
from typing import Tuple

import h5py
import numpy
from silx.utils.retry import RetryError

logger = logging.getLogger(__name__)


def axis_label(name: str, units: str):
    return f"{name} ({units})"


def get_axis_data(nxdata: h5py.Group, name: str) -> Tuple[str, numpy.ndarray]:
    dset = nxdata[name]
    assert isinstance(dset, h5py.Dataset)
    units = dset.attrs["units"]
    assert isinstance(units, str)

    return axis_label(name, units), dset[()]


def find_pyFAI_generated_process(parent: h5py.Group) -> str:
    for grp in parent.values():
        if (
            isinstance(grp, h5py.Group)
            and "program" in grp
            and grp["program"][()].decode() == "pyFAI"
        ):
            return grp.name

    raise KeyError(
        f"Could not find a pyFAI-generated process in {parent.file.filename}::{parent.name}"
    )


def get_integrated_nxdata(root: h5py.File, scan) -> h5py.Group:
    scan_dsetname = f'/{scan.scan_info.get("scan_nb")}.1'

    try:
        scan_dset = root[scan_dsetname]
        nxprocess = find_pyFAI_generated_process(scan_dset)

        nxdata = root[f"{nxprocess}/integrated"]
    except KeyError as e:
        raise RetryError(str(e))
    assert isinstance(nxdata, h5py.Group)

    return nxdata


def nfs_cache_refresh(dirname: str) -> None:
    # _ = os.listdir(dirname)  # not enough
    # _ = subprocess.run(["ls", "-l", dirname], check=False, text=True, capture_output=True)
    try:
        with os.scandir(dirname) as it:
            for entry in it:
                _ = entry.stat()  # This forces a fresh stat call, similar to "ls -l"
    except Exception:
        pass



import logging
import warnings
from typing import Any
from typing import Dict
from typing import Optional

from .persistent.parameters import WithPersistentParameters
from .utils.classes import NoMethodAssignment

logger = logging.getLogger(__name__)


class BlissAutomationObject(WithPersistentParameters, NoMethodAssignment):
    """A Bliss object for automation (data processing and/or acquisition).

    It has persistent parameters in Redis and configuration in Beacon.

    The Beacon configuration is expected to use the
    `generic plugin <https://bliss.gitlab-pages.esrf.fr/bliss/master/config_plugins.html#plugin-generic>`_.

    For example:

    .. code-block:: yaml

        - name: my_processor
            plugin: generic
            class: MyProcessor
            package: blissoda.mycategory.my_processor
            default_parameters:
              param1: value1
              param2: value2

    ``config["default_parameters"]`` has priority over `defaults`.
    """

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        config = config or {}
        defaults = defaults or {}
        beacon_defaults = config.get("default_parameters", {})
        defaults = {**defaults, **beacon_defaults}

        super().__init__(**defaults)

        self._config = config

    @classmethod
    def _merge_defaults(
        cls,
        deprecated_defaults: Dict[str, Any],
        defaults: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        if deprecated_defaults:
            warning = f"{cls.__name__}(**defaults) is deprecated, use {cls.__name__}(defaults={{...}})"
            logger.warning(warning)
            warnings.warn(warning, DeprecationWarning, stacklevel=2)

        if defaults:
            return {**deprecated_defaults, **defaults}
        else:
            return deprecated_defaults



from contextlib import contextmanager
from datetime import datetime

try:
    from bliss.shell.standard import print_html
except ImportError:
    print_html = None


MESSAGE_STYLES = {
    "assert": {
        "passed": ("<ansicyan>[PASSED {timestamp}] {msg}</ansicyan>", "[PASSED] {msg}"),
        "failed": ("<orange>[FAILED {timestamp}] {msg}</orange>", "[FAILED] {msg}"),
    },
    "test": {
        "passed": (
            "<ansigreen>[PASSED {timestamp}] {msg}</ansigreen>",
            "[PASSED] {msg}",
        ),
        "failed": ("<ansired>[FAILED {timestamp}] {msg}</ansired>", "[FAILED] {msg}"),
    },
    "warning": {
        "passed": ("<orange>[WARN {timestamp}] {msg}</orange>", "[WARN] {msg}"),
        "failed": ("<ansired>[FAILED {timestamp}] {msg}</ansired>", "[FAILED] {msg}"),
    },
    "info": {
        "passed": ("<ansicyan>[INFO {timestamp}] {msg}</ansicyan>", "[INFO] {msg}"),
        "failed": ("<ansired>[FAILED {timestamp}] {msg}</ansired>", "[FAILED] {msg}"),
    },
}


@contextmanager
def print_message_on_exit(message: str, message_type: str, indent: int = 0):
    try:
        yield
    except Exception:
        print_message(message, message_type, status="failed")
        raise
    print_message(message, message_type)


def print_message(msg: str, message_type: str, status: str = "passed") -> None:
    try:
        style = MESSAGE_STYLES[message_type][status]
    except KeyError:
        raise ValueError(f"Unknown message_type={message_type!r} or status={status!r}")

    if message_type in ("assert", "info"):
        prefix = " "
    else:
        prefix = ""

    # Time stamps are useful when comparing with Celery time stamps for debugging
    timestamp = datetime.now().strftime("%H:%M:%S,%f")[:-3]

    html_fmt, plain_fmt = style
    if print_html:
        print_html(prefix + html_fmt.format(msg=msg, timestamp=timestamp))
    else:
        print(prefix + plain_fmt.format(msg=msg, timestamp=timestamp))



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy


@dataclass
class XasPlotData:
    name: str
    x: numpy.ndarray
    y: numpy.ndarray
    xlabel: str
    ylabel: str
    info: Dict[str, Any]
    hlines: Sequence[float]
    vlines: Sequence[float]


@dataclass
class XasSubscanData:
    mu: Optional[XasPlotData] = None
    chi: Optional[XasPlotData] = None
    ft: Optional[XasPlotData] = None
    noise: Optional[XasPlotData] = None


@dataclass
class SubScanInfo:
    legend: str
    color: Tuple[int, int, int]
    enabled: bool = True
    updated: bool = False


@dataclass
class ScanInfo:
    filename: str
    scan_number: int
    subscans: List[SubScanInfo]
    xas_results: List[XasSubscanData]
    plot_future: Optional[Any] = None
    split_future: Optional[Any] = None
    reprocess_all: bool = False
    multi_xas_scan: bool = False
    multi_xas_subscan_size: int = 0

    @property
    def scan_url(self) -> str:
        return f"silx://{self.filename}::/{self.scan_number}.1"

    @property
    def min_subscan_index_to_process(self) -> int:
        if not self.multi_xas_scan or self.reprocess_all or not self.xas_results:
            return 0
        return len(self.xas_results) - 1

    @property
    def job_id(self) -> Optional[str]:
        if self.future is None:
            return
        return self.future.uuid


@dataclass
class ExafsPlotWorkflowParameters:
    workflow: str
    energy_name: str
    energy_unit: str
    mu_name: str
    trim_n_points: int = 0


@dataclass
class ExafsSplitWorkflowParameters:
    workflow: dict
    monotonic_channel: str
    scan_complete: bool
    trim_n_points: int = 0



"""Template for new integration tests."""

from .. import testing


def template_demo(expo=0.2, npoints=10):
    for _ in range(2):
        template_test(expo=expo, npoints=npoints)


@testing.integration_fixture
def template_fixture1():
    print("setup 'template_fixture1'")
    yield "value1"
    print("teardown 'template_fixture1' always called, also when test fails")


@testing.integration_fixture
def template_fixture2(template_fixture1):
    print("execute 'template_fixture2'")
    assert template_fixture1 == "value1"
    return "value2"


@testing.integration_test
def template_test(template_fixture1, template_fixture2, expo=0.2, npoints=10):
    assert template_fixture1 == "value1"
    assert template_fixture2 == "value2"



import json
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ...bliss_globals import current_session
from ...bliss_globals import setup_globals
from ...persistent.parameters import ParameterInfo
from ...processor import BlissScanType
from ...utils import directories
from ...xrpd.processor import XrpdProcessor
from ..calib import DEFAULT_CALIB


class DemoXrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("config_filename", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("lima_names", ["difflab6"])
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

        super().__init__(config=config, defaults=defaults)
        if self._HAS_BLISS:
            self._ensure_config_filename()

        self.queue = "celery"

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        self._ensure_config_filename()
        inputs = super().get_integrate_inputs(scan, lima_name, task_identifier)
        is_even = not bool(
            setup_globals.difflab6.image.width % 2
        )  # lima-camera-simulator<1.9.10 does not support odd image widths
        inputs.append(
            {"task_identifier": task_identifier, "name": "demo", "value": is_even}
        )
        return inputs

    def _ensure_config_filename(self):
        if self.config_filename:
            return
        root_dir = self._get_config_dir(current_session.scan_saving.filename)
        cfgfile = os.path.join(root_dir, "pyfaicalib.json")
        os.makedirs(os.path.dirname(cfgfile), exist_ok=True)
        poni = DEFAULT_CALIB
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.config_filename = cfgfile

    def _get_demo_result_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "demo", "xrpd")

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "workflows")

    def _get_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "config")

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.config_filename

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        return self.integration_options.to_dict()



import h5py
import silx.io
from silx.gui.plot import Plot1D
from silx.io.utils import h5py_read_dataset as read_dataset

from ..flint import BasePlot
from ..flint import capture_errors


class BlissIntensityPlots(BasePlot):
    WIDGET = "blissoda.id01.plots.IntensityPlots"

    def clear_data(self):
        self.submit("clear")

    def set_data(self, output_url: str):
        self.submit("set_data", output_url)


class IntensityPlots(Plot1D):
    @capture_errors
    def set_data(self, output_url: str) -> None:
        with silx.io.open(output_url) as entry:
            process = entry["cdi_fits"]
            x0 = process["offset_fit/y"]
            parameters = process["parameters"]
            assert isinstance(parameters, h5py.Group)
            x_label: str = read_dataset(parameters["axis_name"])
            y_label: str = read_dataset(parameters["counter"])
            norm = read_dataset(parameters["norm"]) if "norm" in parameters else None

            for i, nxdata in enumerate(process["intensity_fits"].values()):
                signal = nxdata[nxdata.attrs["signal"]]
                label = signal.attrs["long_name"]
                x = nxdata[nxdata.attrs["axes"][0]]
                fit = nxdata[nxdata.attrs["auxiliary_signals"][0]]

                self.addCurve(x[()], signal[()], legend=f"{label} - {i}")
                self.addCurve(
                    x[()],
                    fit[()],
                    legend=f"fit - {label} - {i}",
                    color="r",
                    linestyle="--",
                )
                self.addCurve(
                    [x0[i], x0[i]], [signal[()].min(), signal[()].max()], legend=str(i)
                )
        self.setGraphXLabel(x_label)
        self.setGraphYLabel(y_label)
        self.setYAxisLogarithmic(norm == "log")


class BlissOffsetPlot(BasePlot):
    WIDGET = "blissoda.id01.plots.OffsetPlot"

    def clear_data(self):
        self.submit("clear")

    def set_data(self, output_url: str):
        self.submit("set_data", output_url)


class OffsetPlot(Plot1D):
    @capture_errors
    def set_data(self, output_url: str) -> None:
        with silx.io.open(output_url) as entry:
            nxdata = entry["cdi_fits/offset_fit"]
            signal = nxdata[nxdata.attrs["signal"]]
            x = nxdata[nxdata.attrs["axes"][0]]
            fit = nxdata[nxdata.attrs["auxiliary_signals"][0]]
            self.addCurve(x[()], signal[()], symbol="o", legend="original data")
            self.addCurve(x[()], fit[()], color="r", legend="fitted line")



import time

from ewoksjob.client import Future
from ewoksjob.client.celery.utils import get_not_finished_futures

from ._display import print_message


def wait_workflows(
    *futures: Future, timeout: int = 30, raise_on_error: bool = False
) -> None:
    if not futures:
        futures = get_not_finished_futures()

    t0 = time.time()
    for future in futures:
        timepassed = time.time() - t0
        try:
            _ = future.result(timeout=timeout - timepassed)
        except Exception as ex:
            if raise_on_error:
                raise
            print_message(f"Workflow failed: {ex}", "warning")
        print_message(f"Job {future.uuid!r} finished", "info")



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from pydantic import BaseModel

from . import defaults


class XrfMapPaths(BaseModel):
    filename: str
    output_root_uri: str
    convert_destination: str
    workflow_path: str
    config_filenames: List[str]


class MosaicXrfMapPaths(BaseModel):
    filenames: List[str]
    output_root_uri: str
    concat_bliss_scan_uri: str
    convert_destination: str
    workflow_path: str
    config_filenames: List[str]


class FluoXasPaths(BaseModel):
    filenames: List[str]
    output_root_uri: str
    convert_destination: str
    workflow_path: str
    config_filenames: List[str]


class CommonParameters(BaseModel):
    instrument_name: Optional[str]
    detector_names: List[str]
    detector_normalization_template: str
    counter_name: Optional[str]
    counter_normalization_template: Optional[str]
    config_filenames: List[str]
    energy_name: str
    fast_fitting: bool
    quantification: bool
    diagnostics: bool
    no_fit: bool
    fit_single_detector: Optional[bool]
    sum_spectra: Optional[bool]
    norm_identifier: str
    fit_identifier: str
    virtual_axes: Optional[Dict[str, str]]
    ignore_axes: Optional[List[str]]
    axis_units: Optional[Dict[str, str]]


class XrfMapParameters(CommonParameters):
    filename: str
    scan_number: int
    output_root_uri: str
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]]


class MosaicXrfMapParameters(CommonParameters):
    filenames: List[str]
    scan_ranges: Sequence[Tuple[int, int]]
    exclude_scans: Sequence[Sequence[int]]
    output_root_uri: str
    bliss_scan_uri: str
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]]


class FluoXasParameters(CommonParameters):
    filenames: List[str]
    scan_ranges: Sequence[Tuple[int, int]]
    exclude_scans: Sequence[Sequence[int]]
    output_root_uri: str
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]]
    align_counter: Optional[str]
    align_crop: Optional[bool]
    fast_align_counter_selection: Optional[bool]


def common_parameters_model(
    detector_numbers: Optional[Sequence[int]],
    detector_names: Optional[Sequence[str]],
    config_filenames: Sequence[str],
    energy_name: Optional[str],
    counter_name: Optional[str],
    instrument_name: Optional[str],
    fast_fitting: bool,
    quantification: bool,
    diagnostics: bool,
    livetime_ref_value: Union[str, int, float, None],
    counter_ref_value: Union[str, int, float, None],
    stack: bool,
    virtual_axes: Optional[Dict[str, str]],
    ignore_axes: Optional[List[str]],
    axis_units: Optional[Dict[str, str]],
) -> Tuple[Sequence[dict], Optional[bool]]:
    detector_names = detector_names or [
        defaults.MCA_NAME_FORMAT[instrument_name].format(i) for i in detector_numbers
    ]
    counter_name = counter_name or defaults.I0_COUNTER[instrument_name]
    if counter_name:
        counter_normalization_template = f"{counter_ref_value or defaults.DEFAULT_COUNTER_REF_VALUE}/<instrument/{{}}/data>"
    else:
        counter_normalization_template = None
    energy_name = energy_name or defaults.ENERGY_COUNTER[instrument_name]
    detector_normalization_template = f"{livetime_ref_value or defaults.DEFAULT_LIVETIME_REF_VALUE}/<instrument/{{}}/live_time>"

    if len(detector_names) == 0 or len(config_filenames) == 0:
        no_fit = True
        sum_spectra = None
        fit_single_detector = None
    elif len(detector_names) == 1:
        # Only one detector
        if len(config_filenames) != 1:
            raise ValueError("Only one pymca configuration is needed for one detector")
        no_fit = False
        sum_spectra = None
        fit_single_detector = True
    elif len(config_filenames) == 1:
        # More than one detector and fit the sum
        no_fit = False
        sum_spectra = True
        fit_single_detector = True
    else:
        # More than one detector and fit each detector separately
        if len(detector_names) != len(config_filenames):
            raise ValueError(
                f"{len(detector_names)} pymca configurations are needed for {len(detector_names)} detectors"
            )
        no_fit = False
        sum_spectra = False
        fit_single_detector = False

    if stack:
        norm_identifier = "NormalizeXrfResultsStack"
        if fit_single_detector:
            fit_identifier = "FitStackSingleDetector"
        else:
            fit_identifier = "FitStackMultiDetector"
    else:
        norm_identifier = "NormalizeXrfResults"
        if fit_single_detector:
            fit_identifier = "FitSingleScanSingleDetector"
        else:
            fit_identifier = "FitSingleScanMultiDetector"

    return CommonParameters(
        instrument_name=instrument_name,
        detector_names=detector_names,
        detector_normalization_template=detector_normalization_template,
        counter_name=counter_name,
        counter_normalization_template=counter_normalization_template,
        energy_name=energy_name,
        fast_fitting=fast_fitting,
        quantification=quantification,
        diagnostics=diagnostics,
        no_fit=no_fit,
        fit_single_detector=fit_single_detector,
        config_filenames=config_filenames,
        sum_spectra=sum_spectra,
        norm_identifier=norm_identifier,
        fit_identifier=fit_identifier,
        virtual_axes=virtual_axes,
        ignore_axes=ignore_axes,
        axis_units=axis_units,
    )



from typing import List
from typing import Optional

from marshmallow import fields

from ..daiquiri.mixin import DaiquiriProcessorMixin
from ..daiquiri.mixin import ExposedParameter
from ..daiquiri.validators import exists_valid_json
from ..resources import resource_filename
from .xrpd_processor import Id13XrpdProcessor


class DaiquiriXrpdProcessor(Id13XrpdProcessor, DaiquiriProcessorMixin):
    DEFAULT_WORKFLOW: Optional[str] = resource_filename("id13", "daiquiri_juno.json")

    EXPOSED_PARAMETERS = [
        ExposedParameter(
            parameter="current_workflow",
            field_type=fields.Str,
            title="Last triggered workflow",
            field_options={"validate": exists_valid_json},
        ),
        ExposedParameter(
            parameter="pyfai_config",
            field_type=fields.Str,
            title="PyFAI Config File",
            field_options={"validate": exists_valid_json},
        ),
        ExposedParameter(
            parameter="do_diffmap",
            field_type=fields.Bool,
            title="Enable Diffmap",
        ),
        ExposedParameter(
            parameter="average_reference",
            field_type=fields.Str,
            title="Name of the reference (standard)",
        ),
        ExposedParameter(
            parameter="do_average",
            field_type=fields.Bool,
            title="Enable Average",
        ),
        ExposedParameter(
            parameter="do_stackedf",
            field_type=fields.Bool,
            title="Enable StackEDF (XRDUA)",
        ),
        ExposedParameter(
            parameter="do_background_removal",
            field_type=fields.Bool,
            title="Background Removal (SPI)",
        ),
        ExposedParameter(
            parameter="do_cnmf",
            field_type=fields.Bool,
            title="CNMF (SPI)",
        ),
        ExposedParameter(
            parameter="do_phase_inference",
            field_type=fields.Bool,
            title="Phase Inference (SPI)",
        ),
        ExposedParameter(
            parameter="integration_options",
            field_type=fields.Dict,
            field_options={"keys": fields.Str()},
            title="PyFAI Options",
        ),
    ]

    def scan_requires_processing(self, scan) -> bool:
        """Only execute this workflow if there is a `daiquiri_datacollectionid` in the `scan_info`"""
        requires_processing = super().scan_requires_processing(scan) and bool(
            scan.scan_info.get("daiquiri_datacollectionid")
        )
        return requires_processing

    def get_daiquiri_inputs(self, scan, lima_name):
        task_identifier = "StartJob"
        return [
            {
                "task_identifier": task_identifier,
                "name": "dataCollectionId",
                "value": scan.scan_info.get("daiquiri_datacollectionid"),
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": self.master_output_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "workflow",
                "value": self.get_workflow(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "config",
                "value": {
                    "pyfai_config": self.get_config_filename(lima_name),
                    "integrate_options": self.get_integration_options(scan, lima_name),
                },
            },
        ]

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        """Merge in daiquiri StartJob options"""
        inputs = super().get_inputs(scan, lima_name)
        inputs += self.get_daiquiri_inputs(scan, lima_name)
        return inputs

    def get_submit_arguments(self, scan, lima_name) -> dict:
        """Need ppf for error handling"""
        args = super().get_submit_arguments(scan, lima_name)
        args["engine"] = "ppf"
        return args



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit
from ewoksutils.task_utils import task_inputs

from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..utils.directories import get_dataset_processed_dir
from ..utils.directories import get_filename
from ..utils.directories import workflow_destination
from .cdi_plotter import CdiPlotter
from .cdi_uploader import CdiUploader


class CdiProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("counter", category="workflows"),
        ParameterInfo("upload_figures_to_logbook", category="ICAT"),
        ParameterInfo("retry_timeout", category="data access"),
        ParameterInfo("retry_period", category="data access"),
        ParameterInfo("queue", category="data processing", validator=str),
    ],
):

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        # https://gitlab.esrf.fr/bliss/blissoda/-/merge_requests/329#note_441492
        defaults = self._merge_defaults(deprecated_defaults, defaults)
        defaults.setdefault("counter", "mpx1x4_roi1")
        defaults.setdefault("upload_figures_to_logbook", False)
        defaults.setdefault("retry_timeout", 60)
        defaults.setdefault("retry_period", 1)
        defaults.setdefault("queue", "online")
        super().__init__(config=config, defaults=defaults)
        # Needed for scan sequence
        self._set_parameter("trigger_at", "END")
        self._plotter = CdiPlotter(max_plots=10)
        self._uploader = CdiUploader()

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[dict]:
        if not scan.scan_info.get("run_ewoks"):
            return
        workflow = {
            "graph": {"id": "cdi_plot"},
            "nodes": [
                {
                    "id": "cdi_plot_task",
                    "task_type": "class",
                    "task_identifier": "ewoksid01.tasks.cdi_plot.CdiPlot",
                }
            ],
        }

        future = submit(
            args=(workflow,),
            kwargs={
                "inputs": self._get_inputs(scan),
                "convert_destination": workflow_destination(scan),
            },
            queue=self.queue,
        )

        self._plotter.handle_workflow_result(
            future,
            self._get_parameter("retry_timeout"),
            self._get_parameter("retry_period"),
        )

        if self._get_parameter("upload_figures_to_logbook"):
            self._uploader.upload_workflow_result(
                future,
                self._get_parameter("retry_timeout"),
                self._get_parameter("retry_period"),
            )

    def _get_inputs(self, sequence: BlissScanType) -> List[Dict[str, Any]]:

        scan_filename = get_filename(sequence)
        return task_inputs(
            inputs={
                "filename": scan_filename,
                "scan_nos": [scan.scan_info.get("scan_nb") for scan in sequence.scans],
                "axis_name": sequence.scan_info["axis_name"],
                "rc_axis_name": sequence.scan_info["rc_axis_name"],
                "counter": self._get_parameter("counter"),
                "output_folder": get_dataset_processed_dir(scan_filename),
                "retry_timeout": self._get_parameter("retry_timeout"),
                "retry_period": self._get_parameter("retry_period"),
            }
        )



from __future__ import annotations

import json
import re
from collections import OrderedDict
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from ewoksjob.client import submit

from ...flint.access import WithFlintAccess
from ...import_utils import unavailable_function
from ...persistent.parameters import ParameterInfo
from ...processor import BaseProcessor

try:
    from bliss.shell.getval import getval_yes_no
except ImportError as ex:
    getval_yes_no = unavailable_function(ex)

from .grid_plotter import GridPlot

try:
    from bliss import current_session
except ImportError:
    current_session = None


MATERIAL_TABLE = {
    # Aluminum
    "al": ("Al", 2.70),
    "aluminum": ("Al", 2.70),
    "aluminium": ("Al", 2.70),
    # Carbon family (choose the right one by key)
    "c": ("C", 2.26),
    "carbon": ("C", 2.26),
    "graphite": ("C", 2.26),
    # Glassy carbon (rods/blocks)
    "glassy carbon": ("C", 1.52),
    "glassy_carbon": ("C", 1.52),
    "glassy_carbon_bloc": ("C", 1.52),
    "glassy carbon bloc": ("C", 1.52),
    "glassy_carbon_block": ("C", 1.52),
    "glassy carbon block": ("C", 1.52),
    # Ambiguous “carbon block/bloc” — keep as graphite unless explicitly “glassy”
    "carbon bloc": ("C", 2.26),
    "carbon block": ("C", 2.26),
    # Carbon fibers (PAN-based typical bulk density)
    "carbon_fibers_5um": ("C", 1.80),
    "carbon ni fibers": ("C", 1.80),
    "carbon_ni_fibers": ("C", 1.80),
    # Beryllium
    "be": ("Be", 1.85),
    "beryllium": ("Be", 1.85),
    # Silicon
    "si": ("Si", 2.3296),
    "silicon": ("Si", 2.3296),
    # Silica / quartz
    "sio2": ("SiO2", 2.20),
    "silica": ("SiO2", 2.20),
    "silica b": ("SiO2", 2.20),
    "silica c": ("SiO2", 2.20),
    "sio2 b": ("SiO2", 2.20),
    "sio2 block": ("SiO2", 2.20),
    "sio2_block": ("SiO2", 2.20),
    "quartz": ("SiO2", 2.65),
    # Alumina / sapphire
    "al2o3": ("Al2O3", 3.98),
    "sapphire": ("Al2O3", 3.98),
    # Copper
    "cu": ("Cu", 8.96),
    "copper": ("Cu", 8.96),
    # Silver
    "ag": ("Ag", 10.49),
    "silver": ("Ag", 10.49),
    # Nickel
    "ni": ("Ni", 8.90),
    "nickel": ("Ni", 8.90),
    # Molybdenum
    "mo": ("Mo", 10.28),
    "molybdenum": ("Mo", 10.28),
    # Titanium
    "ti": ("Ti", 4.506),
    "titanium": ("Ti", 4.506),
    # Zirconium
    "zr": ("Zr", 6.52),
    "zirconium": ("Zr", 6.52),
    # Chromium
    "cr": ("Cr", 7.19),
    "chromium": ("Cr", 7.19),
    # Iron
    "fe": ("Fe", 7.874),
    "iron": ("Fe", 7.874),
    # Tungsten
    "w": ("W", 19.25),
    "tungsten": ("W", 19.25),
    # Gold
    "au": ("Au", 19.32),
    "gold": ("Au", 19.32),
    # Alloys / mixtures
    "brass": ("CuZn", 8.50),
    "lead_brass": ("Cu57Zn40.5Pb2.5", 8.70),
    "lead brass": ("Cu57Zn40.5Pb2.5", 8.70),
    # Polymers / resin (generic)
    "resin": ("C", 1.20),
    "resin lenses": ("C", 1.20),
    "resin_block": ("C", 1.20),
    "resin block": ("C", 1.20),
    # Air (dry, ~15°C, sea level)
    "air": ("Air", 0.001225),
    "free pass": ("Air", 0.001225),
}
_POS = re.compile(r"^\s*([A-Za-z]+)\s*([0-9]+(?:\.[0-9]+)?)\s*$")
_EMPTY = re.compile(
    r"^\s*(empty|out|open|none|0|free\s*pass(?:\s*\d+)?)\s*$", re.IGNORECASE
)
_ORDER_KEY_RE = re.compile(r"^(fix|att)\d+$")

_SYNONYMS = {
    "glassy carbon bloc": "glassy_carbon_bloc",
    "glassy_carbon_bloc": "glassy_carbon_bloc",
    "glassy carbon block": "glassy_carbon_block",
    "glassy_carbon_block": "glassy_carbon_block",
    "glassy carbon": "glassy carbon",
    "glassy_carbon": "glassy carbon",
    "carbon bloc": "carbon bloc",
    "carbon block": "carbon block",
    "carbon fibers 5um": "carbon_fibers_5um",
    "carbon_fibers_5um": "carbon_fibers_5um",
    "carbon ni fibers": "carbon_ni_fibers",
    "carbon_ni_fibers": "carbon_ni_fibers",
    "graphite": "graphite",
    "carbon": "carbon",
    "c": "c",
    "aluminum": "al",
    "aluminium": "al",
    "al": "al",
    "copper": "cu",
    "cu": "cu",
    "silica": "sio2",
    "silica b": "sio2",
    "silica c": "sio2",
    "quartz": "sio2",
    "sio2": "sio2",
    "sio": "sio2",
    "sio2 b": "sio2",
    "sio2 block": "sio2",
    "sio2_block": "sio2",
    "beryllium": "be",
    "be": "be",
    "silicon": "si",
    "si": "si",
    "sapphire": "al2o3",
    "al2o3": "al2o3",
    "silver": "ag",
    "ag": "ag",
    "gold": "au",
    "au": "au",
    "tungsten": "w",
    "w": "w",
    "nickel": "ni",
    "ni": "ni",
    "molybdenum": "mo",
    "mo": "mo",
    "titanium": "ti",
    "ti": "ti",
    "zirconium": "zr",
    "zr": "zr",
    "chromium": "cr",
    "cr": "cr",
    "iron": "fe",
    "fe": "fe",
    "brass": "brass",
    "lead_brass": "lead_brass",
    "lead brass": "lead_brass",
    "resin": "resin",
    "resin lenses": "resin",
    "resin_block": "resin",
    "resin block": "resin",
    "empty": "air",
    "air": "air",
    "free pass": "air",
}
_DESC_NUM = re.compile(r"([0-9]+(?:[.,][0-9]+)?)\s*(mm|um|µm|μm|cm|m)\b", re.IGNORECASE)
_UNIT_MM = {"mm": 1.0, "um": 1e-3, "µm": 1e-3, "μm": 1e-3, "cm": 10.0, "m": 1000.0}


def _to_mm(value: str, unit: Optional[str]) -> float:
    v = float(value.replace(",", "."))
    u = (unit or "").strip().lower()
    return v * _UNIT_MM[u] if u in _UNIT_MM else v


def _material_from_description(desc: str) -> Optional[str]:
    s = desc.lower()
    for k in sorted(_SYNONYMS.keys(), key=len, reverse=True):
        if re.search(rf"(?<![0-9A-Za-z]){re.escape(k)}(?![0-9A-Za-z])", s):
            return _SYNONYMS[k]
    return None


def _parse_description(desc: str) -> Optional[Tuple[str, float]]:
    if not desc:
        return None
    s = re.sub(r"\([^)]*\)", " ", desc)
    s = re.sub(r"\s+", " ", s).strip()
    num = _DESC_NUM.search(s)
    if not num:
        return None
    val, unit = num.groups()
    t_mm = _to_mm(val, unit)
    mat = _material_from_description(s)
    if mat in ("air", None):
        return None
    return mat, float(t_mm)


def _parse_from_positions_list(dev: Any) -> Optional[Tuple[str, float]]:
    plist = getattr(dev, "positions_list", None)
    label = str(getattr(dev, "position", "")).strip().lower()
    if not plist or not label:
        return None
    entry = next(
        (it for it in plist if str(it.get("label", "")).strip().lower() == label), None
    )
    if not entry:
        return None
    got = _parse_description(entry.get("description") or "")
    if not got:
        return None
    mat_key, t_mm = got
    if mat_key not in MATERIAL_TABLE:
        return None
    return mat_key, t_mm


def parse_position_string(position: Any) -> Tuple[Optional[str], float]:
    s = str(position)
    if _EMPTY.match(s):
        return None, 0.0
    m = _POS.match(s)
    if not m:
        raise ValueError(f"Unrecognized attenuator position format: {position!r}")
    code = m.group(1).lower()
    mm = m.group(2)
    thickness_mm = float(mm) if "." in mm else float(int(mm)) / 1000.0
    return code, thickness_mm


class BMEnergyPlotter(WithFlintAccess):
    """Reusable base for plotters that draw in Flint."""

    def __init__(self, *, unique_name: Optional[str] = None) -> None:
        super().__init__()
        self.unique_name = unique_name or self.__class__.__name__
        self._grid: Optional[GridPlot] = None

    @property
    def grid(self) -> GridPlot:
        if self._grid is None:
            self._grid = GridPlot(unique_name=self.unique_name)
        return self._grid

    def clear(self) -> None:
        if self._grid is not None:
            self._grid.clear()

    def _plot_in_flint(self, result: Dict[str, Any]) -> None:
        energy_eV = result.get("energy_eV")
        sp_src = result.get("spectral_power")
        sp_att = result.get("attenuated_spectral_power")
        flux_src = result.get("flux")
        flux_att = result.get("attenuated_flux")
        transmission = result.get("transmission")
        cumulated_power = result.get("cumulated_power")

        x_keV = energy_eV / 1e3
        g = self.grid
        g.set_layout(2, 2)

        # 1) Spectral power
        series_sp = [{"y": sp_src, "label": "Source", "color": "red"}]
        series_sp.append({"y": sp_att, "label": "Attenuated", "color": "green"})
        g.set_cell(
            0,
            0,
            x=x_keV,
            series=series_sp,
            title="Spectral Power (source (red) vs attenuated (green))",
            xlabel="Energy [keV]",
            ylabel="Spectral Power [W/eV]",
        )

        # 2) Flux
        series_flux = []
        series_flux.append(
            {"y": flux_src, "label": "Flux (source)", "color": "darkRed"}
        )
        series_flux.append(
            {"y": flux_att, "label": "Flux (attenuated)", "color": "darkGreen"}
        )
        g.set_cell(
            0,
            1,
            x=x_keV,
            series=series_flux,
            title="Flux (source (red) vs attenuated (green))",
            xlabel="Energy [keV]",
            ylabel="Flux [Phot/s/0.1%bw]",
        )

        # 3) Transmission
        g.set_cell(
            1,
            0,
            x=x_keV,
            series=[{"y": transmission, "label": "T", "color": "blue"}],
            title="Transmission vs Energy",
            xlabel="Energy [keV]",
            ylabel="Transmission",
        )

        # 4) Cumulated power
        g.set_cell(
            1,
            1,
            x=x_keV,
            series=[{"y": cumulated_power, "label": "Cumulated", "color": "black"}],
            title="Cumulated Power vs Energy",
            xlabel="Energy [keV]",
            ylabel="Cumulated Power [W]",
        )

    def plot(self, result: Dict[str, Any]) -> None:
        self._plot_in_flint(result)


class BMEnergyCalculation(
    BaseProcessor,
    parameters=[
        ParameterInfo("attenuators_names", category="attenuators"),
        ParameterInfo("fixed_elements", category="attenuators"),
        ParameterInfo("order", category="attenuators"),
        ParameterInfo("queue", category="workflows"),
        ParameterInfo("prompt_set_energy", category="handle_result"),
        ParameterInfo("plot", category="plot"),
        ParameterInfo("beam_energy_gev", category="source_configuration"),
        ParameterInfo("bfield_t", category="source_configuration"),
        ParameterInfo("current_a", category="source_configuration"),
        ParameterInfo("hor_div_mrad", category="source_configuration"),
        ParameterInfo("phot_energy_min", category="source_configuration"),
        ParameterInfo("phot_energy_max", category="source_configuration"),
        ParameterInfo("npoints", category="source_configuration"),
        ParameterInfo("log_choice", category="source_configuration"),
    ],
):
    """Submit the **BM** workflow and (optionally) plot in Flint via a plotter."""

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ):
        defaults = self._merge_defaults(deprecated_defaults, defaults)
        defaults.update(
            {
                "attenuators_names": "att1,att2,att3,att4,att5",
                "queue": "online",
                "order": None,
                "beam_energy_gev": 6.0,
                "bfield_t": 0.8,
                "current_a": 0.2,
                "hor_div_mrad": 1.0,
                "phot_energy_min": 100.0,
                "phot_energy_max": 200000.0,
                "npoints": 2000,
                "log_choice": 1,
            }
        )
        BaseProcessor.__init__(self, config=config, defaults=defaults)
        self._plotter = BMEnergyPlotter(unique_name="Energy Calculation")

    def _info_categories(self) -> dict:
        cats = super()._info_categories()
        cats.pop("status", None)
        cats.pop("enabled", None)
        return cats

    def submit(self, save_workflow_to: Optional[str] = None):
        wf = self.get_workflow()
        submit_kwargs = self.get_submit_arguments()
        if save_workflow_to:
            p = Path(save_workflow_to)
            p.parent.mkdir(parents=True, exist_ok=True)
            p.write_text(json.dumps(wf, indent=2))
            submit_kwargs["convert_destination"] = save_workflow_to
        future = submit(args=(wf,), kwargs=submit_kwargs, queue=self.queue)
        result = future.get()

        if self.plot and self._plotter is not None:
            self._plotter._plot_in_flint(result)

        mean_eV = self._extract_mean_energy_eV(result)
        if self.prompt_set_energy:
            if mean_eV is None:
                print("Calculated energy is: N/A.")
            else:
                mean_keV = mean_eV / 1e3
                if getval_yes_no(
                    f"Calculated energy is: {mean_keV:.3f} keV. Run ENERGY({mean_keV:.3f})? ",
                    default="no",
                ):
                    self._set_energy_eV(mean_eV)
        else:
            if mean_eV is not None:
                print(f"Calculated energy is: {mean_eV / 1e3:.3f} keV.")
        return

    def get_submit_arguments(self) -> dict:
        return {"inputs": self._build_submit_inputs(), "outputs": [{"all": True}]}

    def get_workflow(self) -> dict:
        return {
            "graph": {"id": "bm_attenuation_bm", "schema_version": "1.1"},
            "nodes": [
                {
                    "id": "compute",
                    "label": "compute_bm_spectrum",
                    "task_type": "class",
                    "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                },
                {
                    "id": "attenuate",
                    "label": "apply_attenuators",
                    "task_type": "class",
                    "task_identifier": "ewokstomo.tasks.energycalculation.ApplyAttenuators",
                },
                {
                    "id": "stats",
                    "label": "spectrum_stats",
                    "task_type": "class",
                    "task_identifier": "ewokstomo.tasks.energycalculation.SpectrumStats",
                },
            ],
            "links": [
                {
                    "source": "compute",
                    "target": "attenuate",
                    "data_mapping": [
                        {"source_output": "energy_eV", "target_input": "energy_eV"},
                        {
                            "source_output": "spectral_power",
                            "target_input": "spectral_power",
                        },
                        {"source_output": "flux", "target_input": "flux"},
                    ],
                },
                {
                    "source": "attenuate",
                    "target": "stats",
                    "data_mapping": [
                        {"source_output": "energy_eV", "target_input": "energy_eV"},
                        {
                            "source_output": "attenuated_flux",
                            "target_input": "attenuated_flux",
                        },
                    ],
                },
            ],
        }

    def _resolve_devices(self) -> List[Any]:
        if current_session is None:
            raise RuntimeError("BLISS current_session is not available.")
        names = [n.strip() for n in str(self.attenuators_names).split(",") if n.strip()]
        sg = getattr(current_session, "setup_globals", current_session)
        return [getattr(sg, name) for name in names]

    def _build_attenuators_from_devices(
        self, devices: Iterable[Any]
    ) -> Tuple[Dict[str, Dict[str, float]], List[str]]:
        attenuators, order_keys = OrderedDict(), []
        for idx, dev in enumerate(devices, 1):
            parsed = _parse_from_positions_list(dev)
            code, t_mm = parsed if parsed else (None, 0.0)
            if (not code) or t_mm <= 0.0:
                element, density, t_mm = "Air", MATERIAL_TABLE["air"][1], 0.0
            else:
                element, density = MATERIAL_TABLE.get(code, (code.capitalize(), 1.0))
            key = f"att{idx}"
            order_keys.append(key)
            attenuators[key] = {
                "material": element,
                "thickness_mm": float(t_mm),
                "density_g_cm3": float(density),
            }
        return attenuators, order_keys

    def _build_submit_inputs(self) -> List[dict]:
        devices = self._resolve_devices()
        att_layers, _ = self._build_attenuators_from_devices(devices)
        fixed_layers = self._build_fixed_elements()

        merged = OrderedDict()
        merged.update(fixed_layers)
        merged.update(att_layers)

        order_value = list(merged.keys()) if self.order is None else self.order

        inputs: List[dict] = [
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "TYPE_CALC",
                "value": 0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "VER_DIV",
                "value": 0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "MACHINE_NAME",
                "value": "ESRF bending magnet",
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "RB_CHOICE",
                "value": 0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "MACHINE_R_M",
                "value": 25.0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "BFIELD_T",
                "value": float(self.bfield_t),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "BEAM_ENERGY_GEV",
                "value": float(self.beam_energy_gev),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "CURRENT_A",
                "value": float(self.current_a),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "HOR_DIV_MRAD",
                "value": float(self.hor_div_mrad),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "PHOT_ENERGY_MIN",
                "value": float(self.phot_energy_min),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "PHOT_ENERGY_MAX",
                "value": float(self.phot_energy_max),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "NPOINTS",
                "value": int(self.npoints),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "LOG_CHOICE",
                "value": int(self.log_choice),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "PSI_MRAD_PLOT",
                "value": 1.0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "PSI_MIN",
                "value": -1.0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "PSI_MAX",
                "value": 1.0,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "PSI_NPOINTS",
                "value": 500,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeBMSpectrum",
                "name": "FILE_DUMP",
                "value": False,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ApplyAttenuators",
                "name": "attenuators",
                "value": merged,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ApplyAttenuators",
                "name": "order",
                "value": order_value,
            },
        ]
        return inputs

    @staticmethod
    def _coerce_layer(entry: Dict[str, Any]) -> Dict[str, float]:
        material = str(entry.get("material", "")).strip()
        if not material:
            raise ValueError("fixed_elements entry missing 'material'")
        thickness_mm = float(entry.get("thickness_mm", 0.0))
        if "density_g_cm3" in entry:
            density = float(entry["density_g_cm3"])
        else:
            mkey = material.lower()
            density = MATERIAL_TABLE.get(mkey, (material, 1.0))[1]

        # Treat empty/air with <=0 mm as neutral layer
        if material.lower() in ("empty", "none", "air") and thickness_mm <= 0.0:
            material = "Air"
            thickness_mm = 0.0
            density = MATERIAL_TABLE["air"][1]

        return {
            "material": material,
            "thickness_mm": thickness_mm,
            "density_g_cm3": density,
        }

    def _build_fixed_elements(self) -> OrderedDict[str, Dict[str, float]]:
        fixed_src = dict(self.fixed_elements or {})
        fixed: "OrderedDict[str, Dict[str, float]]" = OrderedDict()
        gen_idx = 1
        for user_key, entry in fixed_src.items():
            layer = self._coerce_layer(entry)
            key = (
                user_key
                if isinstance(user_key, str) and _ORDER_KEY_RE.match(user_key)
                else f"fix{gen_idx}"
            )
            gen_idx += 1 if not (_ORDER_KEY_RE.match(str(user_key) or "")) else 0
            fixed[key] = layer
        return fixed

    @staticmethod
    def _extract_mean_energy_eV(result: dict) -> Optional[float]:
        v = (result.get("stats") or {}).get(
            "mean_energy_eV", result.get("mean_energy_eV")
        )
        return float(v) if v is not None else None

    def _set_energy_eV(self, energy_eV: float) -> None:
        if current_session is None:
            return
        sg = getattr(current_session, "setup_globals", current_session)
        ENERGY = getattr(sg, "ENERGY", None)
        if callable(ENERGY):
            ENERGY(float(energy_eV) / 1e3)



from __future__ import annotations

from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ...persistent.parameters import ParameterInfo
from .bm_energy_calculation import BMEnergyCalculation


class WigglerEnergyCalculation(
    BMEnergyCalculation,
    parameters=[
        ParameterInfo("attenuators_names", category="attenuators"),
        ParameterInfo("fixed_elements", category="attenuators"),
        ParameterInfo("order", category="attenuators"),
        ParameterInfo("queue", category="workflows"),
        ParameterInfo("prompt_set_energy", category="handle_result"),
        ParameterInfo("plot", category="plot"),
        ParameterInfo("beam_energy_gev", category="source_configuration"),
        ParameterInfo("current_a", category="source_configuration"),
        ParameterInfo("phot_energy_min", category="source_configuration"),
        ParameterInfo("phot_energy_max", category="source_configuration"),
        ParameterInfo("npoints", category="source_configuration"),
        ParameterInfo("log_choice", category="source_configuration"),
        ParameterInfo("wig_field", category="source_configuration"),  # FIELD
        ParameterInfo("wig_nperiods", category="source_configuration"),  # NPERIODS
        ParameterInfo("wig_ulambda_m", category="source_configuration"),  # ULAMBDA (m)
        ParameterInfo("wig_k", category="source_configuration"),  # K
        ParameterInfo(
            "wig_ntrajpoints", category="source_configuration"
        ),  # NTRAJPOINTS
        ParameterInfo("wig_file", category="source_configuration"),  # FILE (dump) or ""
        ParameterInfo("wig_slit_flag", category="source_configuration"),  # SLIT_FLAG
        ParameterInfo("wig_slit_d_m", category="source_configuration"),  # SLIT_D (m)
        ParameterInfo("wig_slit_ny", category="source_configuration"),  # SLIT_NY
        ParameterInfo(
            "wig_slit_width_h_mm", category="source_configuration"
        ),  # SLIT_WIDTH_H_MM
        ParameterInfo(
            "wig_slit_height_v_mm", category="source_configuration"
        ),  # SLIT_HEIGHT_V_MM
        ParameterInfo(
            "wig_slit_center_h_mm", category="source_configuration"
        ),  # SLIT_CENTER_H_MM
        ParameterInfo(
            "wig_slit_center_v_mm", category="source_configuration"
        ),  # SLIT_CENTER_V_MM
        ParameterInfo(
            "wig_shift_x_flag", category="source_configuration"
        ),  # SHIFT_X_FLAG
        ParameterInfo(
            "wig_shift_x_value_m", category="source_configuration"
        ),  # SHIFT_X_VALUE (m)
        ParameterInfo(
            "wig_shift_betax_flag", category="source_configuration"
        ),  # SHIFT_BETAX_FLAG
        ParameterInfo(
            "wig_shift_betax_value", category="source_configuration"
        ),  # SHIFT_BETAX_VALUE (rad)
        ParameterInfo(
            "wig_traj_resampling_factor", category="source_configuration"
        ),  # TRAJ_RESAMPLING_FACTOR
        ParameterInfo(
            "wig_slit_points_factor", category="source_configuration"
        ),  # SLIT_POINTS_FACTOR
    ],
):
    """Submit the **Wiggler** workflow and (optionally) plot in Flint.
    Inherits common logic from BMEnergyCalculation (devices, fixed elements, plotting).
    """

    _HIDE_PARAMS = {"trigger_at", "Enabled"}

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ):
        defaults = self._merge_defaults(deprecated_defaults, defaults)
        defaults.update(
            {
                "attenuators_names": "att1,att2,att3,att4,att5",
                "queue": "online",
                "order": None,
                "beam_energy_gev": 6.0,  # GeV
                "current_a": 0.2,  # A (converted to mA for the task)
                "phot_energy_min": 100.0,  # eV
                "phot_energy_max": 800000.0,  # eV
                "npoints": 4000,
                "log_choice": 1,
                "wig_field": 1,
                "wig_nperiods": 1,
                "wig_ulambda_m": 0.15,
                "wig_k": 22.591,
                "wig_ntrajpoints": 101,
                "wig_file": "",
                "wig_slit_flag": 1,
                "wig_slit_d_m": 56.5,
                "wig_slit_ny": 101,
                "wig_slit_width_h_mm": 10.0,
                "wig_slit_height_v_mm": 5.0,
                "wig_slit_center_h_mm": 0.0,
                "wig_slit_center_v_mm": 0.0,
                "wig_shift_x_flag": 1,
                "wig_shift_x_value_m": -0.002385,
                "wig_shift_betax_flag": 5,
                "wig_shift_betax_value": 0.005,
                "wig_traj_resampling_factor": 10000.0,
                "wig_slit_points_factor": 3.0,
            }
        )
        super().__init__(config=config, defaults=defaults)

    def get_workflow(self) -> dict:
        """Wire ComputeWigglerSpectrum -> ApplyAttenuators -> SpectrumStats."""
        return {
            "graph": {"id": "bm_attenuation_wiggler", "schema_version": "1.1"},
            "nodes": [
                {
                    "id": "compute",
                    "label": "compute_wiggler_spectrum",
                    "task_type": "class",
                    "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                },
                {
                    "id": "attenuate",
                    "label": "apply_attenuators",
                    "task_type": "class",
                    "task_identifier": "ewokstomo.tasks.energycalculation.ApplyAttenuators",
                },
                {
                    "id": "stats",
                    "label": "spectrum_stats",
                    "task_type": "class",
                    "task_identifier": "ewokstomo.tasks.energycalculation.SpectrumStats",
                },
            ],
            "links": [
                {
                    "source": "compute",
                    "target": "attenuate",
                    "data_mapping": [
                        {"source_output": "energy_eV", "target_input": "energy_eV"},
                        {
                            "source_output": "spectral_power",
                            "target_input": "spectral_power",
                        },
                        {"source_output": "flux", "target_input": "flux"},
                    ],
                },
                {
                    "source": "attenuate",
                    "target": "stats",
                    "data_mapping": [
                        {"source_output": "energy_eV", "target_input": "energy_eV"},
                        {
                            "source_output": "attenuated_spectral_power",
                            "target_input": "attenuated_spectral_power",
                        },
                        {
                            "source_output": "attenuated_flux",
                            "target_input": "attenuated_flux",
                        },
                    ],
                },
            ],
        }

    def _build_submit_inputs(self) -> List[dict]:
        devices = self._resolve_devices()
        att_layers, _ = self._build_attenuators_from_devices(devices)
        fixed_layers = self._build_fixed_elements()

        merged = OrderedDict()
        merged.update(fixed_layers)
        merged.update(att_layers)

        order_value = (
            list(merged.keys()) if getattr(self, "order", None) is None else self.order
        )

        inputs: List[dict] = [
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "PHOT_ENERGY_MIN",
                "value": float(self.phot_energy_min),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "PHOT_ENERGY_MAX",
                "value": float(self.phot_energy_max),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "NPOINTS",
                "value": int(self.npoints),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "ENERGY",
                "value": float(self.beam_energy_gev),
            },  # GeV
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "CURRENT",
                "value": float(self.current_a) * 1000.0,
            },  # mA
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "FIELD",
                "value": int(self.wig_field),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "NPERIODS",
                "value": int(self.wig_nperiods),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "ULAMBDA",
                "value": float(self.wig_ulambda_m),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "K",
                "value": float(self.wig_k),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "NTRAJPOINTS",
                "value": int(self.wig_ntrajpoints),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "FILE",
                "value": str(self.wig_file),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_FLAG",
                "value": int(self.wig_slit_flag),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_D",
                "value": float(self.wig_slit_d_m),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_NY",
                "value": int(self.wig_slit_ny),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_WIDTH_H_MM",
                "value": float(self.wig_slit_width_h_mm),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_HEIGHT_V_MM",
                "value": float(self.wig_slit_height_v_mm),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_CENTER_H_MM",
                "value": float(self.wig_slit_center_h_mm),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_CENTER_V_MM",
                "value": float(self.wig_slit_center_v_mm),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SHIFT_X_FLAG",
                "value": int(self.wig_shift_x_flag),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SHIFT_X_VALUE",
                "value": float(self.wig_shift_x_value_m),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SHIFT_BETAX_FLAG",
                "value": int(self.wig_shift_betax_flag),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SHIFT_BETAX_VALUE",
                "value": float(self.wig_shift_betax_value),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "TRAJ_RESAMPLING_FACTOR",
                "value": float(self.wig_traj_resampling_factor),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "SLIT_POINTS_FACTOR",
                "value": float(self.wig_slit_points_factor),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ComputeWigglerSpectrum",
                "name": "LOG_CHOICE",
                "value": int(self.log_choice),
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ApplyAttenuators",
                "name": "attenuators",
                "value": merged,
            },
            {
                "task_identifier": "ewokstomo.tasks.energycalculation.ApplyAttenuators",
                "name": "order",
                "value": order_value,
            },
        ]
        return inputs



import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ..persistent.parameters import ParameterInfo
from ..resources import resource_filename
from ..utils.pyfai import read_config
from ..xrpd.processor import BlissScanType
from ..xrpd.processor import XrpdProcessor
from ..xrpd.processor import _get_scan_memory_url

logger = logging.getLogger(__name__)

SLURM_JOB_PARAMETERS_INTEGRATE = {
    "name": "ID13_pyFAI-integrate",
    "partition": "gpu",
    "time": "01:00:00",
    "tasks_per_node": 1,
    "cpus_per_task": 1,
    "memory_per_cpu": "50G",
    "tres_per_job": "gres/gpu:1",
    "constraints": "l40s",
}

SLURM_JOB_PARAMETERS_BACKGROUND = {
    "name": "ID13_spi-background",
    "partition": "nice",
    "time": "02:00:00",
    "tasks_per_node": 1,
    "cpus_per_task": 32,
    "memory_per_cpu": "2G",
    "tres_per_job": None,
    "constraints": None,
}

SLURM_JOB_PARAMETERS_CNMF = {
    "name": "ID13_spi-CNMF",
    "partition": "gpu",
    "time": "01:00:00",
    "tasks_per_node": 1,
    "cpus_per_task": 1,
    "memory_per_cpu": "50G",
    "tres_per_job": "gres/gpu:1",
    "constraints": "l40s",
}

QUEUE_GPU_ID13 = "lid13gpu3"
QUEUE_SLURM_ID13 = "lid13gpu3_slurm"
WORKER_MODULE = "scattering"
PRE_SCRIPT = "module load {WORKER_MODULE}; python3 -m ewoksid13.scripts.utils.slurm_python_pre_script"
PYTHON_CMD = "python3"
POST_SCRIPT = "python3 -m ewoksid13.scripts.utils.slurm_python_post_script"


class Id13XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("workflow_juno", category="workflows"),
        ParameterInfo("current_workflow", category="workflows"),
        ParameterInfo("queue_slurm", category="workflows"),
        ParameterInfo("normalization_counter", category="Processing"),
        ParameterInfo("do_diffmap", category="Processing"),
        ParameterInfo("do_average", category="Processing"),
        ParameterInfo("do_stackedf", category="Processing"),
        ParameterInfo("average_reference", category="Processing"),
        ParameterInfo("save_external_files", category="Processing"),
        ParameterInfo("radial_limits", category="Processing"),
        ParameterInfo("directory_cif_phases", category="Processing"),
        ParameterInfo("do_background_removal", category="Processing"),
        ParameterInfo("do_cnmf", category="Processing"),
        ParameterInfo("do_phase_inference", category="Processing"),
        ParameterInfo("inference_weights_filename", category="Processing"),
        ParameterInfo("submit_to_slurm_integration", category="Slurm"),
        ParameterInfo("submit_to_slurm_neuralnetwork", category="Slurm"),
    ],
):
    DEFAULT_LIMA_URL_TEMPLATE: Optional[str] = (
        "{dirname}/scan{scan_number_as_str}/{images_prefix}{{file_index}}.h5::/entry_0000/measurement/data"
    )
    DEFAULT_WORKFLOW_JUNO: Optional[str] = resource_filename("id13", "juno.json")

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("current_workflow", "")
        defaults.setdefault("queue_slurm", QUEUE_SLURM_ID13)
        defaults.setdefault("save_scans_separately", True)
        defaults.setdefault("do_diffmap", False)
        defaults.setdefault("normalization_counter", None)
        defaults.setdefault("do_average", False)
        defaults.setdefault("do_stackedf", False)
        defaults.setdefault("average_reference", "hydrocerussite")
        defaults.setdefault("save_external_files", False)
        defaults.setdefault("workflow_juno", self.DEFAULT_WORKFLOW_JUNO)
        defaults.setdefault("do_background_removal", False)
        defaults.setdefault("radial_limits", None)
        defaults.setdefault("do_cnmf", False)
        defaults.setdefault("directory_cif_phases", None)
        defaults.setdefault("do_phase_inference", False)
        defaults.setdefault("inference_weights_filename", "")
        defaults.setdefault("submit_to_slurm_integration", False)
        defaults.setdefault("submit_to_slurm_spi", True)
        super().__init__(config=config, defaults=defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        if self.pyfai_config:
            integration_options = read_config(filename=self.pyfai_config)
        else:
            integration_options = dict()
        if self.integration_options:
            integration_options.update(self.integration_options.to_dict())
        return integration_options

    def get_integrate_inputs(
        self, scan: BlissScanType, lima_name: str, task_identifier: str
    ) -> List[dict]:
        eval_dict = {"img_acq_device": lima_name, "scan_number": scan.scan_number}
        images_prefix = scan.scan_saving.images_prefix.format(**eval_dict)
        original_template_args = self.lima_url_template_args.to_dict()
        self.lima_url_template_args["images_prefix"] = images_prefix
        self.lima_url_template_args["scan_number_as_str"] = scan.scan_number
        try:
            return super().get_integrate_inputs(scan, lima_name, task_identifier)
        finally:
            self.lima_url_template_args = original_template_args

    def get_external_output_filename(self, scan: BlissScanType, lima_name: str):
        external_output_filename = self.external_output_filename(
            scan=scan, lima_name=lima_name
        )
        if not external_output_filename:
            scan_nb = scan.scan_info.get("scan_nb")
            master_output_filename = self.master_output_filename(scan)
            external_output_filename = master_output_filename.replace(
                ".h5", f"_{scan_nb:04d}.h5"
            )
        return external_output_filename

    def get_inputs(self, scan: BlissScanType, lima_name: str) -> List[dict]:
        """Additional inputs for the Id13 XrpdProcessor."""
        inputs = super().get_inputs(scan=scan, lima_name=lima_name)
        inputs += self.get_diff_extra_inputs(scan=scan, lima_name=lima_name)
        inputs += self.get_neuralnetwork_inputs(scan=scan, lima_name=lima_name)
        return inputs

    def get_diff_extra_inputs(
        self,
        scan: BlissScanType,
        lima_name: str,
    ) -> List[dict]:
        """Get additional inputs for the diffmap, average, and stacked EDF tasks."""
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        master_output_filename = self.master_output_filename(scan)
        external_output_filename = self.get_external_output_filename(
            scan=scan, lima_name=lima_name
        )
        external_output_filename_edf = external_output_filename.replace(".h5", ".edf")

        if self.save_external_files:
            external_output_filename = external_output_filename
        else:
            external_output_filename = master_output_filename

        params = {
            "CreateDiffMapFile": {
                "master_filename": filename,
                "scan_nb": scan_nb,
                "output_filename": master_output_filename,
                "external_output_filename": external_output_filename,
                "integration_options": self.get_integration_options(
                    scan=scan, lima_name=lima_name
                ),
                "lima_name": lima_name,
                "do_diffmap": self.do_diffmap,
                "scan_memory_url": _get_scan_memory_url(scan),
                "normalization_counter": self.normalization_counter,
                "processing_info": self.get_saving_info(scan),
            },
            "AverageIntegration": {
                "master_filename": filename,
                "scan_nb": scan_nb,
                "lima_name": lima_name,
                "output_filename": master_output_filename,
                "external_output_filename": external_output_filename,
                "do_average": self.do_average,
                "reference": self.average_reference,
                "scan_memory_url": _get_scan_memory_url(scan),
                "normalization_counter": self.normalization_counter,
                "processing_info": self.get_saving_info(scan),
            },
            "StackToEdf": {
                "do_stackedf": self.do_stackedf,
                "external_output_filename": external_output_filename_edf,
                "processing_info": self.get_saving_info(scan),
            },
        }

        return [
            {"name": name, "value": value, "task_identifier": task_id}
            for task_id, task_params in params.items()
            for name, value in task_params.items()
        ]

    def get_neuralnetwork_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        """Get additional inputs for the neural network (spi) tasks: background removal, constrained NMF and phase inference."""
        external_output_filename = self.get_external_output_filename(
            scan=scan, lima_name=lima_name
        )

        params = {
            "BackgroundRemoval": {
                "wavelength": self.get_wavelength(angstroms=True),
                "radial_limits": self.radial_limits,
                "force_training": False,
                "use_neuralnetwork": False,
                "do_background_removal": self.do_background_removal,
                "submit_to_slurm": self.submit_to_slurm_neuralnetwork,
                "destination_file": external_output_filename.replace(
                    ".h5", "_background.json"
                ),
                "slurm_job_parameters": SLURM_JOB_PARAMETERS_BACKGROUND,
                "worker_module": "scattering",
                "wait_to_finish": self.do_cnmf,
                "slurm_celery_queue": self.queue_slurm,
                "processing_info": self.get_saving_info(scan),
            },
            "ConstrainedNMF": {
                "wavelength": self.get_wavelength(angstroms=True),
                "references_directory": self.directory_cif_phases,
                "do_matrix_factorization": self.do_cnmf,
                "radial_limits": self.radial_limits,
                "submit_to_slurm": self.submit_to_slurm_neuralnetwork,
                "destination_file": external_output_filename.replace(
                    ".h5", "_cnmf.json"
                ),
                "slurm_job_parameters": SLURM_JOB_PARAMETERS_CNMF,
                "worker_module": "scattering",
                "wait_to_finish": False,
                "slurm_celery_queue": self.queue_slurm,
                "processing_info": self.get_saving_info(scan),
            },
            "PhaseInference": {
                "do_phase_inference": self.do_phase_inference,
                "references_directory": self.directory_cif_phases,
                "inference_weights_filename": self.inference_weights_filename,
                "processing_info": self.get_saving_info(scan),
            },
        }

        return [
            {"name": name, "value": value, "task_identifier": task_id}
            for task_id, task_params in params.items()
            for name, value in task_params.items()
        ]

    def get_wavelength(self, angstroms: bool = True) -> float:
        """Load the wavelength from the PyFai configuration file."""
        with open(self.pyfai_config) as f:
            config = json.load(f)
        if "wavelength" in config:
            return float(config["wavelength"]) * (1e10 if angstroms else 1)
        if "poni" in config:
            return float(config["poni"]["wavelength"]) * (1e10 if angstroms else 1)

    def _get_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            self.current_workflow = self.workflow_juno
        else:
            self.current_workflow = self.workflow_without_saving
        return self.current_workflow

    def _set_workflow(self, scan: BlissScanType, filename: str) -> None:
        """Set the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            self.workflow_with_saving = filename
        else:
            self.workflow_without_saving = filename
        self.current_workflow = filename

    def get_submit_arguments(self, scan: BlissScanType, lima_name) -> dict:
        """Additional submit arguments in case of slurm queue, also switch the queue if needed."""
        submit_arguments = super().get_submit_arguments(scan, lima_name)

        if self.submit_to_slurm_integration:
            if self.queue != self.queue_slurm:
                logger.warning(
                    f"Switching queue from {self.queue} to {self.queue_slurm}"
                )
                self.queue = self.queue_slurm
            return {
                **submit_arguments,
                "slurm_arguments": {
                    "parameters": SLURM_JOB_PARAMETERS_INTEGRATE,
                    "pre_script": PRE_SCRIPT.format(WORKER_MODULE=WORKER_MODULE),
                    "python_cmd": PYTHON_CMD,
                    "post_script": POST_SCRIPT,
                },
            }
        return submit_arguments

    def get_saving_info(self, scan: BlissScanType) -> dict:
        return {
            "base_path": scan.scan_saving.base_path,
            "template": scan.scan_saving.template,
            "proposal_dirname": scan.scan_saving.proposal_dirname,
            "beamline": scan.scan_saving.beamline,
            "proposal_session_name": scan.scan_saving.proposal_session_name,
            "collection_name": scan.scan_saving.collection_name,
            "dataset_name": scan.scan_saving.dataset_name,
        }



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .utils import directories
from .utils import models
from .workflows import mosaic_multi_detector
from .workflows import mosaic_multi_detector_sumspectra
from .workflows import mosaic_single_detector


def mosaic_xrfmap_workflow_inputs(
    filenames: Sequence[str],
    output_root_uri: str,
    concat_bliss_scan_uri: str,
    scan_ranges: Sequence[Tuple[int, int]],
    config_filenames: Sequence[str],
    instrument_name: Optional[str] = None,
    detector_numbers: Optional[Sequence[int]] = None,
    detector_names: Optional[Sequence[str]] = None,
    fast_fitting: bool = True,
    quantification: bool = True,
    diagnostics: bool = True,
    livetime_ref_value: Union[str, int, float, None] = None,
    counter_ref_value: Union[str, int, float, None] = None,
    counter_name: Optional[str] = None,
    energy_name: Optional[str] = None,
    exclude_scans: Optional[Sequence[Sequence[int]]] = None,
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]] = None,
    virtual_axes: Optional[Dict[str, str]] = None,
    ignore_axes: Optional[List[str]] = None,
    axis_units: Optional[Dict[str, str]] = None,
) -> Tuple[str, List[dict]]:
    if not exclude_scans:
        exclude_scans = [()] * len(filenames)

    if len(filenames) != len(scan_ranges):
        raise ValueError(
            f"{len(filenames)} scan ranges (first, last) are needed for {len(filenames)} files"
        )

    if len(filenames) != len(exclude_scans):
        raise ValueError(
            f"{len(filenames)} exclude scan list are needed for {len(filenames)} files"
        )

    parameters = models.common_parameters_model(
        detector_numbers,
        detector_names,
        config_filenames,
        energy_name,
        counter_name,
        instrument_name,
        fast_fitting,
        quantification,
        diagnostics,
        livetime_ref_value,
        counter_ref_value,
        False,
        virtual_axes,
        ignore_axes,
        axis_units,
    )

    parameters = models.MosaicXrfMapParameters(
        filenames=filenames,
        scan_ranges=scan_ranges,
        exclude_scans=exclude_scans,
        output_root_uri=output_root_uri,
        bliss_scan_uri=concat_bliss_scan_uri,
        resolution=resolution,
        **parameters.model_dump(),
    )

    if parameters.fit_single_detector:
        if parameters.sum_spectra:
            workflow = mosaic_multi_detector_sumspectra.WORKFLOW
            inputs = mosaic_multi_detector_sumspectra.workflow_inputs(parameters)
        else:
            workflow = mosaic_single_detector.WORKFLOW
            inputs = mosaic_single_detector.workflow_inputs(parameters)
    else:
        workflow = mosaic_multi_detector.WORKFLOW
        inputs = mosaic_multi_detector.workflow_inputs(parameters)

    return workflow, inputs


def mosaic_xrfmap_paths(
    session: str,
    sample: str,
    datasets: Sequence[str],
    scan_ranges: Sequence[Tuple[int, int]],
    config_filenames: Sequence[str],
    dirname: Optional[str] = None,
    demo: bool = False,
) -> models.MosaicXrfMapPaths:
    filenames = [
        directories.raw_directory(session, sample, dataset, demo=demo)
        for dataset in datasets
    ]

    first_dataset = datasets[0]
    first_scan_number = scan_ranges[0][0]

    concat_filename = directories.processed_path(
        session,
        sample,
        first_dataset,
        dirname,
        f"{sample}_{first_dataset}_concat.h5",
        demo=demo,
    )
    output_filename = directories.processed_path(
        session,
        sample,
        first_dataset,
        dirname,
        f"{sample}_{first_dataset}_results.h5",
        demo=demo,
    )
    convert_destination = directories.processed_path(
        session,
        sample,
        first_dataset,
        dirname,
        f"{sample}_{first_dataset}_scan{first_scan_number:04d}.json",
        demo=demo,
    )

    concat_bliss_scan_uri = f"{concat_filename}::/{first_scan_number}.1"
    output_root_uri = f"{output_filename}::/{first_scan_number}.1"
    config_filenames = [
        directories.pymca_config_path(session, s, demo=demo) for s in config_filenames
    ]

    workflow_path = directories.accessible_workflow_path(
        session, sample, first_dataset, dirname, demo=demo
    )

    return models.MosaicXrfMapPaths(
        filenames=filenames,
        output_root_uri=output_root_uri,
        concat_bliss_scan_uri=concat_bliss_scan_uri,
        convert_destination=convert_destination,
        workflow_path=workflow_path,
        config_filenames=config_filenames,
    )



import json
import os
from typing import Any
from typing import Dict
from typing import Optional

from ..bliss_globals import current_session
from ..bm02.xrpd_processor import Bm02XrpdProcessor
from ..utils import directories
from .calib import DEFAULT_CALIB


class DemoBm02XrpdProcessor(Bm02XrpdProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("queue", "celery")
        defaults.setdefault("lima_names", ["difflab6"])
        defaults.setdefault(
            "integration_options",
            {
                "difflab6": {
                    "method": "no_csr_cython",
                    "nbpt_rad": 4096,
                    "unit": "q_A^-1",
                }
            },
        )
        defaults.setdefault(
            "empty_cell_subtraction_options",
            {
                "difflab6": {
                    "enabled": False,
                    "cell_pattern_url": None,
                    "empty_cell_factor": 1,
                }
            },
        )

        super().__init__(config=config, defaults=defaults)

        if self._HAS_BLISS:
            self._ensure_pyfai_config_filename()

    def _ensure_pyfai_config_filename(self):
        root_dir = directories.get_processed_dir(current_session.scan_saving.filename)
        cfgfile = os.path.join(root_dir, "config", "pyfaicalib.json")
        if not os.path.exists(cfgfile):
            poni = DEFAULT_CALIB
            os.makedirs(os.path.dirname(cfgfile), exist_ok=True)
            with open(cfgfile, "w") as f:
                json.dump(poni, f)
        self.config_filename = {"difflab6": cfgfile}


bm02_xrpd_processor = DemoBm02XrpdProcessor()



import json
import logging
import pickle
from collections import OrderedDict
from collections.abc import MutableMapping
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Union

_ERROR = None
try:
    from blissdata.settings import OrderedHashObjSetting
except ImportError:
    try:
        from bliss.config.settings import OrderedHashObjSetting
    except ImportError as ex:
        _ERROR = ex

        class OrderedHashObjSetting(MutableMapping):
            def __init__(self, name: str):
                logger.warning(
                    "Persistency is disabled because of this error: %s", _ERROR
                )
                self.__data: Dict[Any, Any] = OrderedDict()

            def __getitem__(self, key: Any) -> Any:
                return self.__data[key]

            def __setitem__(self, key: Any, value: Any) -> None:
                self.__data[key] = value

            def __delitem__(self, key: Any) -> None:
                del self.__data[key]

            def __iter__(self) -> Iterator[Any]:
                return iter(self.__data)

            def __len__(self) -> int:
                return len(self.__data)

            def get_all(self) -> dict:
                return deepcopy(self._data)


logger = logging.getLogger(__name__)


class PersistentOrderedDict(OrderedHashObjSetting):
    def __init__(self, redis_key: str, serializer: str = "json"):
        dumps, loads = _SERIALIZERS[serializer]
        super().__init__(
            redis_key, read_type_conversion=loads, write_type_conversion=dumps
        )


def _json_loads(serialized_value: Union[bytes, Any]) -> Any:
    # Called twice on the same data for bliss<2
    if isinstance(serialized_value, bytes):
        try:
            return json.loads(serialized_value)
        except Exception:
            pass
    return serialized_value


def _pickle_loads(serialized_value: Union[bytes, Any]) -> Any:
    # Called twice on the same data for bliss<2
    if isinstance(serialized_value, bytes):
        try:
            return pickle.loads(serialized_value)
        except Exception:
            pass
    return serialized_value


def _json_dumps(python_value: Any) -> str:
    return json.dumps(python_value)


def _pickle_dumps(python_value: Any) -> bytes:
    return pickle.dumps(python_value)


_SERIALIZERS = {
    "json": (_json_dumps, _json_loads),
    "pickle": (_pickle_dumps, _pickle_loads),
}



from typing import List

import numpy
from silx.io.url import DataUrl

from ..import_utils import unavailable_type

try:
    from bliss.scanning.scan import Scan as BlissScanType
except ImportError as ex:
    BlissScanType = unavailable_type(ex)


def _gonio2pyFAI(angles: numpy.ndarray) -> numpy.ndarray:
    """Convert goniometer angles (orientation 3: counter-clockwise, origin at beam center, in degree 0-360) to azimuthal angles
    in pyFAI's coordinate system (counter-clockwise, origin at horizontal, radians -pi+pi)
    """
    return numpy.deg2rad(-((angles + 180) % 360))


def get_data_url(scan: BlissScanType, lima_name: str) -> str:
    filename = scan.scan_info.get("filename")
    scan_number = scan.scan_info.get("scan_nb")
    data_url = DataUrl(
        file_path=filename, data_path=f"/{scan_number}.1/measurement/{lima_name}"
    )
    return data_url.path()


def get_positions(start: float, step: float, npts: int) -> List[float]:
    # Shift positions by step / 2 so that bins land in the middle of two positions
    new_start = start + step / 2
    gonio_positions = numpy.arange(new_start, new_start + npts * step, step)
    return _gonio2pyFAI(gonio_positions).tolist()


def is_multigeometry(scan: BlissScanType) -> bool:
    return scan.scan_info.get("type") == "fscan" and scan.scan_info["save"]



from typing import List

from ...id14.converter import Id14Hdf5ToSpecConverter
from ...import_utils import unavailable_type

try:
    from bliss.scanning.scan import Scan as BlissScanType
except ImportError as ex:
    BlissScanType = unavailable_type(ex)


class DemoId14Hdf5ToSpecConverter(Id14Hdf5ToSpecConverter):
    def _get_inputs_for_mca(self, scan: BlissScanType) -> List[dict]:
        inputs = super()._get_inputs_for_mca(scan)
        task_identifier = "Hdf5ToSpec"
        inputs.append(
            {
                "task_identifier": task_identifier,
                "name": "mca_counter",
                "value": "mca1_det0",
            }
        )
        return inputs

    def _scan_requires_mca_conversion(self, scan: BlissScanType) -> bool:
        return True

    def _scan_requires_asc_conversion(self, scan: BlissScanType) -> bool:
        return True



"""Integration tests for the ID14 HDF5 converter."""

import pathlib

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.id14 import DemoId14Hdf5ToSpecConverter

ID14_CONVERTER = DemoId14Hdf5ToSpecConverter()


@testing.integration_fixture
def id14_converter():
    ID14_CONVERTER.enable()
    yield ID14_CONVERTER
    ID14_CONVERTER.disable()


def id14_demo(expo=0.2, npoints=10):
    for _ in range(2):
        test_id14_loopscan(expo=expo, npoints=npoints)


@testing.integration_test
def test_id14_loopscan(id14_converter, expo=0.2, npoints=10):
    scan = setup_globals.loopscan(
        npoints, expo, setup_globals.diode1, setup_globals.mca1
    )
    scan_number = scan.scan_info["scan_nb"]
    _assert_spec_counter_file(id14_converter, scan_number)
    _assert_spec_mca_file(id14_converter, scan_number)


@testing.demo_assert("Check ID14 SPEC file for scan #{scan_number}")
def _assert_spec_counter_file(id14_converter, scan_number):
    result = id14_converter._future_for_counters.result(timeout=10)
    output_filename = pathlib.Path(result["output_filename"])
    testing.assert_spec_scan_exists(output_filename, scan_number)


@testing.demo_assert("Check ID14 MCA file for scan #{scan_number}")
def _assert_spec_mca_file(id14_converter, scan_number):
    result = id14_converter._future_for_mca.result(timeout=10)
    output_filenames = result["output_filenames"]
    if not output_filenames:
        raise AssertionError("No filenames returned from ID14 MCA conversion workflow")
    for output_filename in output_filenames:
        testing.assert_spec_scan_exists(output_filename, scan_number)



import functools
from contextlib import ExitStack
from typing import Any
from typing import Callable

from ._display import print_message_on_exit
from ._fixtures import _DEMO_FIXTURE_AUTOUSE
from ._fixtures import _function_id
from ._fixtures import _resolve_fixture_arguments


def integration_test(fn: Callable[..., Any]) -> Callable[..., Any]:
    if not callable(fn):
        raise TypeError(f"{fn} is not a callable")

    @functools.wraps(fn)
    def wrapper(*args, **kwargs) -> Any:
        with ExitStack() as stack:
            _enter_autouse_fixtures(stack)
            final_args, final_kwargs = _resolve_fixture_arguments(
                fn, stack, args=args, kwargs=kwargs
            )

            with print_message_on_exit(_function_id(fn), "test"):
                return fn(*final_args, **final_kwargs)

    return wrapper


def _enter_autouse_fixtures(stack: ExitStack) -> None:
    for fixture_ctx in _DEMO_FIXTURE_AUTOUSE:
        stack.enter_context(fixture_ctx())



from blissoda.bliss_globals import setup_globals
from blissoda.demo.bm02 import bm02_xrpd_processor


def _scan_1d(expo=0.2, npoints=10):
    if "nbpt_azim" in bm02_xrpd_processor.integration_options["difflab6"]:
        bm02_xrpd_processor.integration_options["difflab6"].pop("nbpt_azim")
    return setup_globals.loopscan(npoints, expo, setup_globals.difflab6)


def bm02_demo_1d(expo=0.2, npoints=10, ascii=False):
    bm02_xrpd_processor.enable(setup_globals.difflab6)
    bm02_xrpd_processor.ascii_export_enabled = ascii
    try:
        _scan_1d(expo, npoints)
    finally:
        bm02_xrpd_processor.disable()


def bm02_demo_2d(expo=0.2, npoints=10, ascii=False):
    bm02_xrpd_processor.enable(setup_globals.difflab6)
    try:
        bm02_xrpd_processor.integration_options["difflab6"]["nbpt_azim"] = 100
        bm02_xrpd_processor.ascii_export_enabled = ascii
        setup_globals.loopscan(npoints, expo, setup_globals.difflab6)
    finally:
        bm02_xrpd_processor.disable()


def bm02_demo_1d_with_cell_subtraction(expo=0.2, npoints=10, ascii=False):
    bm02_xrpd_processor.enable(setup_globals.difflab6)
    bm02_xrpd_processor.ascii_export_enabled = ascii

    try:
        # Do a first scan to be subtracted
        scan = _scan_1d(expo, npoints)
        scan_metadata = scan.scan_saving

        bm02_xrpd_processor.enable_empty_cell_subtraction(setup_globals.difflab6)
        bm02_xrpd_processor.set_cell_pattern_url(
            sample=scan_metadata.sample_name,
            dataset=scan_metadata.dataset,
            scan_number=scan.scan_info["scan_nb"],
            detector="difflab6",
        )

        _scan_1d(expo, npoints)
    finally:
        bm02_xrpd_processor.disable()
